# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to the Privacy app."""

from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext_noop

from plinth import app as app_module
from plinth import menu
from plinth.config import DropinConfigs
from plinth.modules.backups.components import BackupRestore
from plinth.package import Packages

from . import manifest, privileged

_description = [_('Manage system-wide privacy settings.')]


class PrivacyApp(app_module.App):
    """FreedomBox app for Privacy."""

    app_id = 'privacy'

    _version = 2

    can_be_disabled = False

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               is_essential=True, name=_('Privacy'),
                               icon='fa-eye-slash', description=_description,
                               manual_page=None)
        self.add(info)

        menu_item = menu.Menu('menu-privacy', info.name,
                              info.short_description, info.icon,
                              'privacy:index', parent_url_name='system:data',
                              order=10)
        self.add(menu_item)

        packages = Packages('packages-privacy', ['popularity-contest', 'gpg'])
        self.add(packages)

        dropin_configs = DropinConfigs('dropin-configs-privacy', [
            '/etc/dpkg/origins/freedombox',
        ])
        self.add(dropin_configs)

        backup_restore = BackupRestore('backup-restore-privacy',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup()
        if old_version == 0:
            privileged.set_configuration(enable_popcon=True)
            _show_privacy_notification()


def _show_privacy_notification():
    """Show a notification asking user to review privacy settings."""
    from plinth.notification import Notification
    message = gettext_noop(
        'Please update privacy settings to match your preferences.')
    data = {
        'app_name': 'translate:' + gettext_noop('Privacy'),
        'app_icon': 'fa-eye-slash'
    }
    title = gettext_noop('Review privacy setting')
    actions_ = [{
        'type': 'link',
        'class': 'primary',
        'text': gettext_noop('Go to {app_name}'),
        'url': 'privacy:index'
    }, {
        'type': 'dismiss'
    }]
    Notification.update_or_create(id='privacy-review', app_id='privacy',
                                  severity='info', title=title,
                                  message=message, actions=actions_, data=data,
                                  group='admin')



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.utils.translation import gettext_lazy as _

from plinth import cfg
from plinth.utils import format_lazy

clients = [{
    'name':
        _('Gobby'),
    'description':
        _('Gobby is a collaborative text editor'),
    'usage':
        format_lazy(
            _('Start Gobby and select "Connect to Server" and '
              'enter your {box_name}\'s domain name.'),
            box_name=_(cfg.box_name)),
    'platforms': [{
        'type': 'download',
        'os': 'gnu-linux',
        'url': 'https://github.com/gobby/gobby/wiki/Download'
    }, {
        'type': 'download',
        'os': 'windows',
        'url': 'https://github.com/gobby/gobby/wiki/Download'
    }, {
        'type': 'package',
        'format': 'deb',
        'name': 'gobby'
    }]
}]

backup = {
    'data': {
        'directories': ['/var/lib/infinoted/']
    },
    'secrets': {
        'files': [
            '/etc/infinoted/infinoted-cert.pem',
            '/etc/infinoted/infinoted-key.pem'
        ],
    },
    'services': ['infinoted']
}



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to configure samba."""

import grp
import pwd
import socket

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import frontpage, menu
from plinth.daemon import Daemon
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import Firewall
from plinth.modules.users.components import UsersAndGroups
from plinth.package import Packages
from plinth.utils import format_lazy

from . import manifest, privileged

_description = [
    _('Samba allows to share files and folders between FreedomBox and '
      'other computers in your local network.'),
    format_lazy(
        _('After installation, you can choose which disks to use for sharing. '
          'Enabled shares are accessible in the file manager on your computer '
          'at location \\\\{hostname} (on Windows) or smb://{hostname}.local '
          '(on Linux and Mac). There are three types of shares '
          'you can choose from: '), hostname=socket.gethostname()),
    _('Open share - accessible to everyone in your local network.'),
    _('Group share - accessible only to FreedomBox users who are in the '
      'freedombox-share group.'),
    _('Home share - every user in the freedombox-share group can have their '
      'own private space.'),
]


class SambaApp(app_module.App):
    """FreedomBox app for Samba file sharing."""

    app_id = 'samba'

    _version = 5

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        groups = {'freedombox-share': _('Access to the private shares')}

        info = app_module.Info(
            app_id=self.app_id, version=self._version, name=_('Samba'),
            icon_filename='samba', short_description=_('Network File Storage'),
            manual_page='Samba', description=_description,
            clients=manifest.clients,
            donation_url='https://www.samba.org/samba/donations.html')
        self.add(info)

        menu_item = menu.Menu('menu-samba', info.name, info.short_description,
                              info.icon_filename, 'samba:index',
                              parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut(
            'shortcut-samba', info.name,
            short_description=info.short_description, icon=info.icon_filename,
            description=info.description, manual_page=info.manual_page,
            configure_url=reverse_lazy('samba:index'), clients=info.clients,
            login_required=True, allowed_groups=list(groups))
        self.add(shortcut)

        packages = Packages('packages-samba', ['samba', 'acl'])
        self.add(packages)

        firewall = Firewall('firewall-samba', info.name, ports=['samba'])
        self.add(firewall)

        daemon = Daemon(
            'daemon-samba', 'smbd', listen_ports=[(139, 'tcp4'), (139, 'tcp6'),
                                                  (445, 'tcp4'),
                                                  (445, 'tcp6')])
        self.add(daemon)

        users_and_groups = UsersAndGroups('users-and-groups-samba',
                                          groups=groups)
        self.add(users_and_groups)

        backup_restore = SambaBackupRestore('backup-restore-samba',
                                            **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup()
        if not old_version:
            self.enable()

    def uninstall(self):
        """De-configure and uninstall the app."""
        super().uninstall()
        privileged.uninstall()


class SambaBackupRestore(BackupRestore):
    """Component to backup/restore Samba."""

    def backup_pre(self, packet):
        """Save registry share configuration."""
        super().backup_pre(packet)
        privileged.dump_shares()

    def restore_post(self, packet):
        """Restore configuration."""
        super().restore_post(packet)
        privileged.setup()
        privileged.restore_shares()


def add_share(mount_point, share_type, filesystem):
    """Add a share."""
    windows_filesystem = (filesystem in ['ntfs', 'vfat'])
    privileged.add_share(mount_point, share_type, windows_filesystem)


def get_users():
    """Get non-system users who are in the freedombox-share or admin group."""
    samba_users = privileged.get_users()
    group_users = grp.getgrnam('freedombox-share').gr_mem + grp.getgrnam(
        'admin').gr_mem

    allowed_users = []
    for group_user in group_users:
        try:
            uid = pwd.getpwnam(group_user).pw_uid
        except KeyError:  # User doesn't exist
            continue

        if uid > 1000:
            allowed_users.append(group_user)

    return {
        'access_ok':
            sorted(set(allowed_users) & set(samba_users)),
        'password_re_enter_needed':
            sorted(set(allowed_users) - set(samba_users))
    }



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.utils.translation import gettext_lazy as _

from plinth.clients import store_url

clients = [{
    'name':
        _('DAVx5'),
    'usage':
        _('Enter the URL of the Radicale server (e.g. '
          'https://<your.freedombox.address>) and your user name. DAVx5 will '
          'show all existing calendars and address books and you can '
          'create new.'),
    'platforms': [{
        'type': 'store',
        'os': 'android',
        'store_name': 'f-droid',
        'url': store_url('f-droid', 'at.bitfire.davdroid'),
    }, {
        'type': 'store',
        'os': 'android',
        'store_name': 'google-play',
        'url': store_url('google-play', 'at.bitfire.davdroid'),
    }]
}, {
    'name':
        _('GNOME Calendar'),
    'platforms': [{
        'type': 'package',
        'format': 'deb',
        'name': 'gnome-calendar'
    }]
}, {
    'name':
        _('Thunderbird'),
    'platforms': [{
        'type': 'package',
        'format': 'deb',
        'name': 'thunderbird'
    }, {
        'type': 'download',
        'os': 'gnu-linux',
        'url': 'https://www.thunderbird.net/'
    }, {
        'type': 'download',
        'os': 'macos',
        'url': 'https://www.thunderbird.net/'
    }, {
        'type': 'download',
        'os': 'windows',
        'url': 'https://www.thunderbird.net/'
    }]
}, {
    'name':
        _('Evolution'),
    'description':
        _('Evolution is a personal information management '
          'application that provides integrated mail, '
          'calendaring and address book functionality.'),
    'usage':
        _('In Evolution add a new calendar and address book '
          'respectively with WebDAV. Enter the URL of the Radicale '
          'server (e.g. https://<your.freedombox.address>) and your '
          'user name. '
          'Clicking on the search button will list the existing '
          'calendars and address books.'),
    'platforms': [{
        'type': 'package',
        'format': 'deb',
        'name': 'evolution'
    }]
}, {
    'name': _('Radicale'),
    'platforms': [{
        'type': 'web',
        'url': '/radicale/'
    }]
}]

backup = {
    'config': {
        'directories': ['/etc/radicale/']
    },
    'data': {
        'directories': ['/var/lib/radicale/']
    },
    'services': ['uwsgi']
}



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Schedule for automatic backups.

Every day automatic backups are triggered. Daily, weekly and monthly backups
are taken. Cleanup of old backups is triggered and specified number of backups
are kept back in each category.

"""

import json
import logging
from datetime import datetime, timedelta

logger = logging.getLogger(__name__)


class Schedule:
    """Description of a schedule for backups."""

    def __init__(self, repository_uuid, enabled=None, daily_to_keep=None,
                 weekly_to_keep=None, monthly_to_keep=None, run_at_hour=None,
                 unselected_apps=None):
        """Initialize the schedule object instance.

        'repository_uuid' is the unique ID of the repository that this
        schedule is applied to.

        'enabled' is a boolean indicating whether scheduled backups are enabled
        or disabled.

        'daily_to_keep' is a whole number indicating the number of daily
        backups to keep. Older backups are removed after creating a new backup
        to keep these many backups. A value of 0 means no such backups are
        scheduled.

        'weekly_to_keep' is a whole number indicating the number of weekly
        backups to keep. Older backups are removed after creating a new backup
        to keep these many backups. A value of 0 means no such backups are
        scheduled.

        'monthly_to_keep' is a whole number indicating the number of monthly
        backups to keep. Older backups are removed after creating a new backup
        to keep these many backups. A value of 0 means no such backups are
        scheduled.

        'run_at_hour' is a whole number indicating the hour of the day when the
        backups must be scheduled.

        'unselected_apps' is a list of app IDs that should not be included when
        scheduling backups. A negative list is maintained because when a new
        app is installed, it is included into the schedule by default unless
        explicitly removed. This is the safer option.

        """
        self.repository_uuid = repository_uuid
        self.enabled = enabled or False
        self.daily_to_keep = daily_to_keep if daily_to_keep is not None else 5
        self.weekly_to_keep = weekly_to_keep if weekly_to_keep is not None \
            else 3
        self.monthly_to_keep = monthly_to_keep if monthly_to_keep is not None \
            else 3
        # Run at 02:00 by default everyday
        self.run_at_hour = run_at_hour if run_at_hour is not None else 2
        self.unselected_apps = unselected_apps or []

    def get_storage_format(self):
        """Return the object serialized as dict suitable for instantiation."""
        return {
            'enabled': self.enabled,
            'daily_to_keep': self.daily_to_keep,
            'weekly_to_keep': self.weekly_to_keep,
            'monthly_to_keep': self.monthly_to_keep,
            'run_at_hour': self.run_at_hour,
            'unselected_apps': self.unselected_apps,
        }

    @staticmethod
    def _is_backup_too_soon(recent_backup_times):
        """Return whether a backup was already taken recently."""
        now = datetime.now()
        if now - recent_backup_times['daily'] < timedelta(seconds=2 * 3600):
            return True

        if now - recent_backup_times['weekly'] < timedelta(seconds=2 * 3600):
            return True

        if now - recent_backup_times['monthly'] < timedelta(seconds=2 * 3600):
            return True

        return False

    @staticmethod
    def _too_long_since_last_backup(recent_backup_times):
        """Return periods for which it has been too long since last backup."""
        periods = []
        local_time = datetime.now()

        if local_time - recent_backup_times['daily'] > timedelta(
                days=1, seconds=3600):
            periods.append('daily')

        if local_time - recent_backup_times['weekly'] > timedelta(
                days=7, seconds=3600):
            periods.append('weekly')

        last_monthly = recent_backup_times['monthly']
        try:
            next_monthly = last_monthly.replace(month=last_monthly.month + 1)
        except ValueError:
            next_monthly = last_monthly.replace(month=1,
                                                year=last_monthly.year + 1)
        if local_time > next_monthly + timedelta(seconds=3600):
            periods.append('monthly')

        return periods

    def _time_for_periods(self):
        """Return periods for which it is scheduled time for backup."""
        periods = []
        local_time = datetime.now()

        if local_time.hour == self.run_at_hour:
            periods.append('daily')  # At specified hour

        if local_time.hour == self.run_at_hour and local_time.isoweekday(
        ) == 7:
            periods.append('weekly')  # At specified hour on Sunday

        if local_time.hour == self.run_at_hour and local_time.day == 1:
            periods.append('monthly')  # At specified hour on 1st of the month

        return periods

    def _get_disabled_periods(self):
        """Return the list of periods for which backup are disabled."""
        periods = []
        if self.daily_to_keep == 0:
            periods.append('daily')

        if self.weekly_to_keep == 0:
            periods.append('weekly')

        if self.monthly_to_keep == 0:
            periods.append('monthly')

        return periods

    def run_schedule(self):
        """Return scheduled backups, throw exception on failure.

        Frequent triggering: If the method is triggered too frequently for any
        reason, later triggers will not result in more backups as the previous
        backup has to be more than 2 hours old at least to trigger a new
        backup.

        Daemon offline: When the daemon is offline, no backups will be made.
        However, an hour after it is back online, backup check will be done and
        if it is determined that too much time has passed since the last
        backup, a new backup will be taken.

        Errors: When an error occurs during backup process, the method raises
        an exception. An hour later, it is triggered again. This time it will
        determine that too much time has passed since the last backup and will
        attempt to backup again. During a day about 24 attempts to backup will
        be made and reported. A backup may made at an unscheduled time due to
        this. This won't prevent the next backup from happening at the
        scheduled time (unless it is too close to the previous successful one).

        Clock changes: When the clock changes and is set forward, within an
        hour of the change, the schedule check will determine that it has been
        too long since the last backup and the backup will be triggered. This
        will result in a backup at an unscheduled time. When the clock changes
        and is set backward, it will result all previous backup that are more
        recent and current time being ignored for scheduling. Backups will be
        scheduled on time and error handling works.

        Day light saving time: When gaining an hour, it is possible that
        schedule qualifies a second time during a day for backup. This is
        avoided by checking if the backup is too soon since the earlier one.
        When loosing an hour, the schedule may not quality for backup on that
        day at all. However, an hour after scheduled time, it will deemed that
        too much time has passed since the previous backup and a backup will be
        scheduled.

        """
        if not self.enabled:
            return False

        repository = self._get_repository()
        repository.prepare()

        recent_backup_times = self._get_recent_backup_times(repository)
        if self._is_backup_too_soon(recent_backup_times):
            return False

        too_long_periods = self._too_long_since_last_backup(
            recent_backup_times)
        time_for_periods = self._time_for_periods()
        disabled_periods = self._get_disabled_periods()
        periods = set(too_long_periods).union(time_for_periods)
        periods = periods.difference(disabled_periods)
        if not periods:
            return False

        self._run_backup(periods)
        self._run_cleanup(repository)
        return True

    def _get_repository(self):
        """Return the repository to which this schedule is assigned."""
        from . import repository as repository_module
        return repository_module.get_instance(self.repository_uuid)

    @staticmethod
    def _serialize_comment(data):
        """Represent dictionary data as comment.

        Borg substitutes python like placeholders with {}.

        """
        comment = json.dumps(data)
        return comment.replace('{', '{{').replace('}', '}}')

    @staticmethod
    def _list_scheduled_archives(repository):
        """Return a list of archives due to scheduled backups."""
        now = datetime.now()

        archives = repository.list_archives()
        scheduled_archives = []
        for archive in archives:

            try:
                comment = json.loads(archive['comment'])
            except json.decoder.JSONDecodeError:
                continue

            if not isinstance(comment, dict) or \
               comment.get('type') != 'scheduled' or \
               not isinstance(comment.get('periods'), list):
                continue

            archive['comment'] = comment

            start_time = datetime.strptime(archive['start'],
                                           '%Y-%m-%dT%H:%M:%S.%f')
            if start_time > now:
                # This backup was taken when clock was set in future. Ignore it
                # to ensure backups continue to be taken.
                continue

            archive['start'] = start_time
            scheduled_archives.append(archive)

        return scheduled_archives

    def _get_recent_backup_times(self, repository):
        """Get the time since most recent daily, weekly and monthly backups."""
        times = {
            'daily': datetime.min,
            'weekly': datetime.min,
            'monthly': datetime.min
        }

        archives = self._list_scheduled_archives(repository)
        for archive in archives:
            periods = {'daily', 'weekly', 'monthly'}
            periods = periods.intersection(archive['comment']['periods'])
            for period in periods:
                if times[period] < archive['start']:
                    times[period] = archive['start']

        return times

    def _run_backup(self, periods):
        """Run a backup and mark it for given period."""
        logger.info('Running backup for repository %s, periods %s',
                    self.repository_uuid, periods)

        from . import api
        periods = list(periods)
        periods.sort()
        name = 'scheduled: {periods}: {datetime}'.format(
            periods=', '.join(periods),
            datetime=datetime.now().strftime('%Y-%m-%d:%H:%M'))
        comment = self._serialize_comment({
            'type': 'scheduled',
            'periods': periods
        })
        app_ids = [
            component.app_id
            for component in api.get_all_components_for_backup()
            if component.app_id not in self.unselected_apps
        ]

        repository = self._get_repository()
        repository.create_archive(name, app_ids, archive_comment=comment)

    def _run_cleanup(self, repository):
        """Cleanup old backups."""
        archives = self._list_scheduled_archives(repository)
        counts = {'daily': 0, 'weekly': 0, 'monthly': 0}
        for archive in archives:
            keep = False
            archive_periods = archive['comment']['periods']
            for period in set(counts).intersection(archive_periods):
                counts[period] += 1

                if period == 'daily' and counts[period] <= self.daily_to_keep:
                    keep = True

                if period == 'weekly' and \
                   counts[period] <= self.weekly_to_keep:
                    keep = True

                if period == 'monthly' and \
                   counts[period] <= self.monthly_to_keep:
                    keep = True

            if not keep:
                logger.info('Cleaning up in repository %s backup archive %s',
                            self.repository_uuid, archive['name'])
                repository.delete_archive(archive['name'])

        repository.cleanup()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configuration helper for Dynamic DNS."""

import pathlib
import urllib
from typing import Any

from plinth.actions import privileged

_conf_dir = pathlib.Path('/etc/ez-ipupdate/')
_active_config = _conf_dir / 'ez-ipupdate.conf'
_inactive_config = _conf_dir / 'ez-ipupdate.inactive'
_helper_config = _conf_dir / 'ez-ipupdate-plinth.cfg'
_cron_job = pathlib.Path('/etc/cron.d/ez-ipupdate')


def _read_configuration(path, separator='='):
    """Read ez-ipupdate configuration."""
    config = {}
    for line in path.read_text().splitlines():
        if line.startswith('#'):
            continue

        parts = line.partition(separator)
        if parts[1]:
            config[parts[0].strip()] = parts[2].strip()
        else:
            config[parts[0].strip()] = True

    return config


@privileged
def export_config() -> dict[str, bool | dict[str, dict[str, str | None]]]:
    """Return the old ez-ipupdate configuration in JSON format."""
    input_config = {}
    if _active_config.exists():
        input_config = _read_configuration(_active_config)
    elif _inactive_config.exists():
        input_config = _read_configuration(_inactive_config)

    helper = {}
    if _helper_config.exists():
        helper.update(_read_configuration(_helper_config, separator=' '))

    def _clean(value):
        value_map = {'enabled': True, 'disabled': False, '': None}
        return value_map.get(value, value)

    domain = {
        'service_type': 'gnudip',
        'domain': input_config.get('host'),
        'server': input_config.get('server'),
        'username': input_config.get('user', '').split(':')[0] or None,
        'password': input_config.get('user', '').split(':')[-1] or None,
        'ip_lookup_url': helper.get('IPURL'),
        'update_url': _clean(helper.get('POSTURL')) or None,
        'use_http_basic_auth': _clean(helper.get('POSTAUTH')),
        'disable_ssl_cert_check': _clean(helper.get('POSTSSLIGNORE')),
        'use_ipv6': _clean(helper.get('POSTUSEIPV6')),
    }

    if isinstance(domain['update_url'], bool):
        # 'POSTURL ' is a line found in the configuration file
        domain['update_url'] = None

    if not domain['server']:
        domain['service_type'] = 'other'
        update_url = domain['update_url']
        try:
            server = urllib.parse.urlparse(update_url).netloc
            service_types: dict[str, str] = {
                'dynupdate.noip.com': 'noip.com',
                'dynupdate.no-ip.com': 'noip.com',
                'freedns.afraid.org': 'freedns.afraid.org'
            }
            domain['service_type'] = service_types.get(str(server), 'other')
        except ValueError:
            pass

    # Old logic for 'enabling' the app is as follows: If behind NAT, add
    # cronjob. If not behind NAT and type is update URL, add cronjob. If not
    # behind NAT and type is GnuDIP, move inactive configuration to active
    # configuration and start the ez-ipupdate daemon.
    enabled = False
    if _cron_job.exists() or (domain['service_type'] == 'gnudip'
                              and _active_config.exists()):
        enabled = True

    output_config: dict[str, Any] = {'enabled': enabled, 'domains': {}}
    if domain['domain']:
        output_config['domains'][domain['domain']] = domain

    return output_config


@privileged
def clean():
    """Remove all old configuration files."""
    last_update = _conf_dir / 'last-update'
    status = _conf_dir / 'ez-ipupdate.status'
    current_ip = _conf_dir / 'ez-ipupdate.currentIP'

    cleanup_files = [
        _active_config, _inactive_config, last_update, _helper_config, status,
        current_ip
    ]
    for cleanup_file in cleanup_files:
        try:
            cleanup_file.rename(cleanup_file.with_suffix('.bak'))
        except FileNotFoundError:
            pass

    _cron_job.unlink(missing_ok=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app to configure Mumble server.
"""

import pathlib

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import frontpage, menu
from plinth.daemon import Daemon
from plinth.modules import names
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import Firewall
from plinth.modules.letsencrypt.components import LetsEncrypt
from plinth.modules.users.components import UsersAndGroups
from plinth.package import Packages, install
from plinth.utils import Version

from . import manifest, privileged

_description = [
    _('Mumble is an open source, low-latency, encrypted, high quality '
      'voice chat software.'),
    _('You can connect to your Mumble server on the regular Mumble port '
      '64738. <a href="http://mumble.info">Clients</a> to connect to Mumble '
      'from your desktop and mobile devices are available.')
]


class MumbleApp(app_module.App):
    """FreedomBox app for Mumble."""

    app_id = 'mumble'

    _version = 2

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(
            app_id=self.app_id, version=self._version, name=_('Mumble'),
            icon_filename='mumble', short_description=_('Voice Chat'),
            description=_description, manual_page='Mumble',
            clients=manifest.clients,
            donation_url='https://wiki.mumble.info/wiki/Donate')
        self.add(info)

        menu_item = menu.Menu('menu-mumble', info.name, info.short_description,
                              'mumble', 'mumble:index', parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut(
            'shortcut-mumble', info.name,
            short_description=info.short_description, icon=info.icon_filename,
            description=info.description, manual_page=info.manual_page,
            configure_url=reverse_lazy('mumble:index'), clients=info.clients)
        self.add(shortcut)

        packages = Packages('packages-mumble', ['mumble-server'])
        self.add(packages)

        firewall = Firewall('firewall-mumble', info.name,
                            ports=['mumble-plinth'], is_external=True)
        self.add(firewall)

        letsencrypt = LetsEncrypt(
            'letsencrypt-mumble', domains=get_domains,
            daemons=['mumble-server'], should_copy_certificates=True,
            private_key_path='/var/lib/mumble-server/privkey.pem',
            certificate_path='/var/lib/mumble-server/fullchain.pem',
            user_owner='mumble-server', group_owner='mumble-server',
            managing_app='mumble')
        self.add(letsencrypt)

        daemon = Daemon(
            'daemon-mumble', 'mumble-server', listen_ports=[(64738, 'tcp4'),
                                                            (64738, 'tcp6'),
                                                            (64738, 'udp4'),
                                                            (64738, 'udp6')])
        self.add(daemon)

        users_and_groups = UsersAndGroups('users-and-groups-mumble',
                                          reserved_usernames=['mumble-server'])
        self.add(users_and_groups)

        backup_restore = BackupRestore('backup-restore-mumble',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup()
        if not old_version:
            self.enable()

        self.get_component('letsencrypt-mumble').setup_certificates()

    def force_upgrade(self, packages):
        """Force upgrade mumble-server to resolve conffile prompts."""
        if 'mumble-server' not in packages:
            return False

        # Allow upgrades within 1.3.*
        package = packages['mumble-server']
        if Version(package['new_version']) > Version('1.4~'):
            return False

        install(['mumble-server'], force_configuration='new')
        privileged.setup()
        return True


def get_available_domains():
    """Return an iterator with all domains able to have a certificate."""
    return (domain.name for domain in names.components.DomainName.list()
            if domain.domain_type.can_have_certificate)


def get_domain():
    """Read TLS domain from config file select first available if none."""
    domain = privileged.get_domain()
    if not domain:
        domain = next(get_available_domains(), None)
        privileged.set_domain(domain)

    return domain


def get_domains():
    """Return a list with the configured domains."""
    if not pathlib.Path('/var/lib/mumble-server/').exists():
        return []

    domain = get_domain()
    if domain:
        return [domain]

    return []



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for ikiwiki app.
"""

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.ikiwiki]


class TestIkiwikiApp(functional.BaseAppTests):
    app_name = 'ikiwiki'
    has_service = False
    has_web = True

    @pytest.mark.backups
    def test_backup_restore(self, session_browser):
        """Test backup and restore of app data."""
        functional.app_enable(session_browser, 'ikiwiki')
        _create_wiki_if_needed(session_browser)
        functional.backup_create(session_browser, 'ikiwiki', 'test_ikiwiki')

        _delete_wiki(session_browser)
        functional.backup_restore(session_browser, 'ikiwiki', 'test_ikiwiki')

        assert _wiki_exists(session_browser)


def _create_wiki_if_needed(browser):
    """Create wiki if it does not exist."""
    functional.nav_to_module(browser, 'ikiwiki')
    wiki = browser.links.find_by_href('/ikiwiki/wiki')
    if not wiki:
        browser.links.find_by_href(
            '/plinth/apps/ikiwiki/create/').first.click()
        browser.find_by_id('id_ikiwiki-name').fill('wiki')
        browser.find_by_id('id_ikiwiki-admin_name').fill(
            functional.config['DEFAULT']['username'])
        browser.find_by_id('id_ikiwiki-admin_password').fill(
            functional.config['DEFAULT']['password'])
        functional.submit(browser, form_class='form-create')


def _delete_wiki(browser):
    """Delete wiki."""
    functional.nav_to_module(browser, 'ikiwiki')
    browser.links.find_by_href(
        '/plinth/apps/ikiwiki/wiki/delete/').first.click()
    functional.submit(browser, form_class='form-delete')


def _wiki_exists(browser):
    """Check whether the wiki exists."""
    functional.nav_to_module(browser, 'ikiwiki')
    wiki = browser.links.find_by_href('/ikiwiki/wiki')
    return bool(wiki)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app for Apache server."""

import os

from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import cfg
from plinth.config import DropinConfigs
from plinth.daemon import Daemon, RelatedDaemon
from plinth.modules.firewall.components import Firewall
from plinth.modules.letsencrypt.components import LetsEncrypt
from plinth.package import Packages
from plinth.utils import format_lazy, is_valid_user_name

from . import privileged


class ApacheApp(app_module.App):
    """FreedomBox app for Apache web server."""

    app_id = 'apache'

    _version = 12

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               is_essential=True, name=_('Apache HTTP Server'))
        self.add(info)

        packages = Packages('packages-apache', [
            'apache2', 'php-fpm', 'ssl-cert', 'uwsgi', 'uwsgi-plugin-python3'
        ])
        self.add(packages)

        dropin_configs = DropinConfigs('dropin-configs-apache', [
            '/etc/apache2/conf-available/php-fpm-freedombox.conf',
            '/etc/fail2ban/jail.d/apache-auth-freedombox.conf',
        ])
        self.add(dropin_configs)

        web_server_ports = Firewall('firewall-web', _('Web Server'),
                                    ports=['http', 'https'], is_external=True)
        self.add(web_server_ports)

        freedombox_ports = Firewall(
            'firewall-plinth',
            format_lazy(_('{box_name} Web Interface (Plinth)'),
                        box_name=_(cfg.box_name)), ports=['http', 'https'],
            is_external=True)
        self.add(freedombox_ports)

        letsencrypt = LetsEncrypt('letsencrypt-apache', domains='*',
                                  daemons=['apache2'])
        self.add(letsencrypt)

        daemon = Daemon('daemon-apache', 'apache2')
        self.add(daemon)

        related_daemon = RelatedDaemon('related-daemon-apache', 'uwsgi')
        self.add(related_daemon)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup(old_version)
        self.enable()


# (U)ser (W)eb (S)ites


def uws_directory_of_user(user):
    """Return the directory of the given user's website."""
    return '/home/{}/public_html'.format(user)


def uws_url_of_user(user):
    """Return the url path of the given user's website."""
    return '/~{}/'.format(user)


def user_of_uws_directory(directory):
    """Return the user of a given user website directory."""
    if directory.startswith('/home/'):
        pos_ini = 6
    elif directory.startswith('home/'):
        pos_ini = 5
    else:
        return None

    pos_end = directory.find('/public_html')
    if pos_end == -1:
        return None

    user = directory[pos_ini:pos_end]
    return user if is_valid_user_name(user) else None


def user_of_uws_url(url):
    """Return the user of a given user website url path."""
    MISSING = -1

    pos_ini = url.find('~')
    if pos_ini == MISSING:
        return None

    pos_end = url.find('/', pos_ini)
    if pos_end == MISSING:
        pos_end = len(url)

    user = url[pos_ini + 1:pos_end]
    return user if is_valid_user_name(user) else None


def uws_directory_of_url(url):
    """Return the directory of the user's website for the given url path.

    Note: It doesn't return the full OS file path to the url path!
    """
    return uws_directory_of_user(user_of_uws_url(url))


def uws_url_of_directory(directory):
    """Return the url base path of the user's website for the given OS path.

    Note: It doesn't return the url path for the file!
    """
    return uws_url_of_user(user_of_uws_directory(directory))


def get_users_with_website():
    """Return a dictionary of users with actual website subdirectory."""

    def lst_sub_dirs(directory):
        """Return the list of subdirectories of the given directory."""
        return [
            name for name in os.listdir(directory)
            if os.path.isdir(os.path.join(directory, name))
        ]

    return {
        name: uws_url_of_user(name)
        for name in lst_sub_dirs('/home')
        if os.path.isdir(uws_directory_of_user(name))
    }



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure infinoted."""

import grp
import os
import pathlib
import pwd
import shutil
import subprocess
import time

from plinth.actions import privileged

DATA_DIR = '/var/lib/infinoted'
KEY_DIR = '/etc/infinoted'
SYNC_DIR = os.path.join(DATA_DIR, 'sync')

CONF_PATH = '/etc/xdg/infinoted.conf'
CONF = f'''
[infinoted]

# Possible values : no-tls, allow-tls, require-tls
security-policy=require-tls

# Absolute path of the certificate file.
certificate-file=/etc/infinoted/infinoted-cert.pem

# Absolute path of the private key file.
key-file=/etc/infinoted/infinoted-key.pem

# Enable plugins
plugins=note-text;autosave;logging;directory-sync

# Specify a path to use a root certificate instead of a certificate-key pair.
#certificate-chain=

#password=

# Automatically save documents every few seconds
[autosave]

# Setting this to 0 disables autosave.
interval=60

# Synchronize files to another directory in plain text format
[directory-sync]

# Directory to sync plain text files
directory={SYNC_DIR}

# Synchronize seconds
interval=60

# Log additional events
[logging]

# Log when users connect or disconnect
log-connections=true

# Log errors with client connections such as a connection reset
log-connection-errors=true
'''

SYSTEMD_SERVICE_PATH = '/etc/systemd/system/infinoted.service'
SYSTEMD_SERVICE = '''
#
# This file is managed and overwritten by Plinth.  If you wish to edit
# it, disable infinoted in Plinth, remove this file and manage it manually.
#

[Unit]
Description=collaborative text editor service
Documentation=man:infinoted(1)
After=network.target

[Service]
User=infinoted
Group=infinoted
ExecStart=/usr/bin/infinoted
ConfigurationDirectory=infinoted
ConfigurationDirectoryMode=0750
LockPersonality=yes
NoNewPrivileges=yes
PrivateDevices=yes
PrivateMounts=yes
PrivateTmp=yes
ProtectControlGroups=yes
ProtectHome=yes
ProtectKernelLogs=yes
ProtectKernelModules=yes
ProtectKernelTunables=yes
ProtectSystem=full
RestrictAddressFamilies=AF_UNIX AF_INET AF_INET6
RestrictRealtime=yes
StateDirectory=infinoted
SystemCallArchitectures=native

[Install]
WantedBy=multi-user.target
'''


def _kill_daemon():
    """Try to kill the infinoted daemon for upto 5 minutes."""
    end_time = time.time() + 300
    while time.time() < end_time:
        try:
            subprocess.run(['infinoted', '--kill-daemon'], check=True)
            break
        except subprocess.CalledProcessError:
            pass

        time.sleep(1)


@privileged
def setup():
    """Configure infinoted after install."""
    if not os.path.isfile(CONF_PATH):
        with open(CONF_PATH, 'w', encoding='utf-8') as file_handle:
            file_handle.write(CONF)

    with open(SYSTEMD_SERVICE_PATH, 'w', encoding='utf-8') as file_handle:
        file_handle.write(SYSTEMD_SERVICE)

    subprocess.check_call(['systemctl', 'daemon-reload'])

    # Create infinoted group if needed.
    try:
        grp.getgrnam('infinoted')
    except KeyError:
        subprocess.run(['addgroup', '--system', 'infinoted'], check=True)

    # Create infinoted user if needed.
    try:
        pwd.getpwnam('infinoted')
    except KeyError:
        subprocess.run([
            'adduser', '--system', '--ingroup', 'infinoted', '--home',
            DATA_DIR, '--gecos', 'Infinoted collaborative editing server',
            'infinoted'
        ], check=True)

    for directory in (DATA_DIR, KEY_DIR, SYNC_DIR):
        if not os.path.exists(directory):
            os.makedirs(directory, mode=0o750)
        shutil.chown(directory, user='infinoted', group='infinoted')

    if not os.path.exists(KEY_DIR + '/infinoted-cert.pem'):
        old_umask = os.umask(0o027)
        try:
            # infinoted doesn't have a "create key and exit" mode. Run as
            # daemon so we can stop after.
            subprocess.run([
                'infinoted', '--create-key', '--create-certificate',
                '--daemonize'
            ], check=True)
            _kill_daemon()
        finally:
            os.umask(old_umask)

    # Always check the ownership of certificate files, in case setup
    # failed previously.
    shutil.chown(KEY_DIR + '/infinoted-cert.pem', user='infinoted',
                 group='infinoted')
    shutil.chown(KEY_DIR + '/infinoted-key.pem', user='infinoted',
                 group='infinoted')


@privileged
def uninstall():
    """Remove data, certs, config and systemd unit files."""
    for directory in DATA_DIR, KEY_DIR:
        shutil.rmtree(directory, ignore_errors=True)

    for file_ in CONF_PATH, SYSTEMD_SERVICE_PATH:
        pathlib.Path(file_).unlink(missing_ok=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Component for managing a background daemon or any systemd unit."""

import contextlib
import socket
import subprocess

import psutil
from django.utils.translation import gettext_noop

from plinth import action_utils, app
from plinth.diagnostic_check import (DiagnosticCheck,
                                     DiagnosticCheckParameters, Result)


class Daemon(app.LeaderComponent):
    """Component to manage a background daemon or any systemd unit."""

    def __init__(self, component_id: str, unit: str,
                 strict_check: bool = False,
                 listen_ports: list[tuple[int, str]] | None = None,
                 alias: str | None = None):
        """Initialize a new daemon component.

        'component_id' must be a unique string across all apps and components
        of a app. Conventionally starts with 'daemon-'.

        'unit' must the name of systemd unit that this component should manage.

        'listen_ports' is a list of tuples. Each tuple contains the port number
        as integer followed by a string with one of the values 'tcp4', 'tcp6',
        'tcp', 'udp4', 'udp6', 'udp' indicating the protocol that the daemon
        listens on. This information is used to run diagnostic tests.

        'alias' is an alternate name for the same unit file. When a unit file
        is renamed, the new unit file usually contains an Alias= setting in
        [Install] section with value of old unit name. When the unit is
        enabled, a symlink with the name of the alias is created. All
        operations such as is-running and disable work with the alias along
        with the primary unit name. However, for the case of enabling the unit
        file or checking its enabled status, the alias does not work. To be
        able to provide management for multiple versions of the unit file with
        different names, specify an alias. Both the names are taken into
        consideration when enabling the unit file.

        """
        super().__init__(component_id)

        self.unit = unit
        self.strict_check = strict_check
        self.listen_ports = listen_ports or []
        self.alias = alias

    def is_enabled(self):
        """Return if the daemon/unit is enabled."""
        if self.alias:
            # XXX: Handling alias should not be done here. service_is_enabled()
            # should return True even for an alias. Currently, in addition to
            # return code we are also checking the printed value. This makes
            # the implementation less future-proof as new values could printed
            # by the command. A fixed systemd bug
            # https://github.com/systemd/systemd/issues/18134 also currently
            # gives incorrect exit code for 'alias' case. See:
            # https://salsa.debian.org/freedombox-team/freedombox/-/merge_requests/1980
            if action_utils.service_is_enabled(self.alias,
                                               strict_check=self.strict_check):
                return True

        return action_utils.service_is_enabled(self.unit,
                                               strict_check=self.strict_check)

    def enable(self):
        """Run operations to enable the daemon/unit."""
        from plinth.privileged import service as service_privileged
        service_privileged.enable(self.unit)
        if self.alias:
            service_privileged.enable(self.alias)

    def disable(self):
        """Run operations to disable the daemon/unit."""
        from plinth.privileged import service as service_privileged
        service_privileged.disable(self.unit)
        if self.alias:
            service_privileged.disable(self.alias)

    def is_running(self):
        """Return whether the daemon/unit is running."""
        return action_utils.service_is_running(self.unit)

    @contextlib.contextmanager
    def ensure_running(self):
        """Ensure a service is running and return to previous state."""
        from plinth.privileged import service as service_privileged
        starting_state = self.is_running()
        if not starting_state:
            service_privileged.enable(self.unit)

        try:
            yield starting_state
        finally:
            if not starting_state:
                service_privileged.disable(self.unit)

    def diagnose(self) -> list[DiagnosticCheck]:
        """Check if the daemon is running and listening on expected ports.

        See :py:meth:`plinth.app.Component.diagnose`.

        """
        results = []
        results.append(self._diagnose_unit_is_running())
        for port in self.listen_ports:
            results.append(
                diagnose_port_listening(port[0], port[1], None,
                                        self.component_id))

        return results

    def _diagnose_unit_is_running(self) -> DiagnosticCheck:
        """Check if a daemon is running."""
        check_id = f'daemon-running-{self.unit}'
        result = Result.PASSED if self.is_running() else Result.FAILED

        description = gettext_noop('Service {service_name} is running')
        parameters: DiagnosticCheckParameters = {
            'service_name': str(self.unit)
        }

        return DiagnosticCheck(check_id, description, result, parameters,
                               self.component_id)


class RelatedDaemon(app.FollowerComponent):
    """Component to hold information about additional systemd units handled.

    Unlike a daemon described by the Daemon component which is enabled/disabled
    when the app is enabled/disabled, the daemon described by this component is
    unaffected by the app's enabled/disabled status. The app only has an
    indirect interest in this daemon.

    This component primarily holds information about such daemon and does
    nothing else. This information is used to check if the app is allowed to
    perform operations on the daemon.
    """

    def __init__(self, component_id, unit):
        """Initialize a new related daemon component.

        'component_id' must be a unique string across all apps and components
        of a app. Conventionally starts with 'related-daemon-'.

        'unit' must the name of systemd unit.

        """
        super().__init__(component_id)

        self.unit = unit


class SharedDaemon(Daemon):
    """Component to manage a daemon that is used by multiple apps.

    Daemons such as a database server are a hard requirement for an app.
    However, there may be multiple apps using that server. This component
    ensures that server is enabled and running when app is enabled. It runs
    diagnostics on the daemon when app is diagnosed. The primary difference
    from the Daemon component is that when the app is disabled the daemon must
    only be disabled if there is no other app using this daemon.
    """

    # A shared daemon may be running even when an app is disabled because
    # another app might be using the daemon. Hence, the enabled/disabled state
    # of this component can't be used to determine the enabled/disabled state
    # of the app.
    is_leader = False

    def set_enabled(self, enabled):
        """Do nothing. Enabled state is still determined by unit status."""

    def disable(self):
        """Disable the daemon iff this is the last app using the daemon."""
        other_apps_enabled = False
        for other_app in app.App.list():
            if other_app.app_id == self.app_id:
                continue

            for component in other_app.get_components_of_type(SharedDaemon):
                if component.unit == self.unit and other_app.is_enabled():
                    other_apps_enabled = True

        if not other_apps_enabled:
            super().disable()


def app_is_running(app_):
    """Return whether all the daemons in the app are running."""
    for component in app_.components.values():
        if hasattr(component, 'is_running') and not component.is_running():
            return False

    return True


def diagnose_port_listening(
        port: int, kind: str = 'tcp',
        listen_address: str | None = None,
        component_id: str | None = None) -> DiagnosticCheck:
    """Run a diagnostic on whether a port is being listened on.

    Kind must be one of inet, inet4, inet6, tcp, tcp4, tcp6, udp,
    udp4, udp6, unix, all.  See psutil.net_connection() for more
    information.

    """
    result = _check_port(port, kind, listen_address)

    parameters: DiagnosticCheckParameters = {'kind': kind, 'port': port}
    if listen_address:
        parameters['listen_address'] = listen_address
        check_id = f'daemon-listening-address-{kind}-{port}-{listen_address}'
        description = gettext_noop(
            'Listening on {kind} port {listen_address}:{port}')
    else:
        check_id = f'daemon-listening-{kind}-{port}'
        description = gettext_noop('Listening on {kind} port {port}')

    return DiagnosticCheck(check_id, description,
                           Result.PASSED if result else Result.FAILED,
                           parameters, component_id)


def _check_port(port: int, kind: str = 'tcp',
                listen_address: str | None = None) -> bool:
    """Return whether a port is being listened on."""
    run_kind = kind

    if kind == 'tcp4':
        run_kind = 'tcp'

    if kind == 'udp4':
        run_kind = 'udp'

    for connection in psutil.net_connections(run_kind):
        # TCP connections must have status='listen'
        if kind in ('tcp', 'tcp4', 'tcp6') and \
           connection.status != psutil.CONN_LISTEN:
            continue

        # UDP connections must have empty remote address
        if kind in ('udp', 'udp4', 'udp6') and \
           connection.raddr != ():
            continue

        # Port should match
        if connection.laddr[1] != port:  # type: ignore[misc]
            continue

        # Listen address if requested should match
        if listen_address and connection.laddr[
                0] != listen_address:  # type: ignore[misc]
            continue

        # Special additional checks only for IPv4
        if kind not in ('tcp4', 'udp4'):
            return True

        # Found socket is IPv4
        if connection.family == socket.AF_INET:
            return True

        # Full IPv6 address range includes mapped IPv4 address also
        if connection.laddr[0] == '::':  # type: ignore[misc]
            return True

    return False


def diagnose_netcat(host: str, port: int, remote_input: str = '',
                    negate: bool = False,
                    component_id: str | None = None) -> DiagnosticCheck:
    """Run a diagnostic using netcat."""
    try:
        process = subprocess.Popen(['nc', host, str(port)],
                                   stdin=subprocess.PIPE,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        process.communicate(input=remote_input.encode())
        if process.returncode != 0:
            result = Result.FAILED if not negate else Result.PASSED
        else:
            result = Result.PASSED if not negate else Result.FAILED
    except Exception:
        result = Result.FAILED

    check_id = f'daemon-netcat-{host}-{port}'
    description = gettext_noop('Connect to {host}:{port}')
    parameters: DiagnosticCheckParameters = {
        'host': host,
        'port': port,
        'negate': negate
    }
    if negate:
        check_id = f'daemon-netcat-negate-{host}-{port}'
        description = gettext_noop('Cannot connect to {host}:{port}')

    return DiagnosticCheck(check_id, description, result, parameters,
                           component_id)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configuration helper for samba."""

import configparser
import os
import pathlib
import shutil
import subprocess

from plinth.actions import privileged

SHARES_CONF_BACKUP_FILE = '/var/lib/plinth/backups-data/samba-shares-dump.conf'
DEFAULT_FILE = '/etc/default/samba'

CONF_PATH = '/etc/samba/smb-freedombox.conf'
CONF = r'''
#
# This file is managed and overwritten by Plinth.  If you wish to manage
# Samba yourself, disable Samba in Plinth, remove this file and remove
# line with --configfile parameter in /etc/default/samba.
#
# Configuration parameters which differ from Debian default configuration
# are commented. To view configured samba shares use command `net conf list`.
#

[global]
   workgroup = WORKGROUP
   log file = /var/log/samba/log.%m
   max log size = 1000
   logging = file
   panic action = /usr/share/samba/panic-action %d
   server role = standalone server
   #obey pam restrictions = yes
   #unix password sync = yes
   #passwd program = /usr/bin/passwd %u
   #passwd chat = *Enter\snew\s*\spassword:* %n\n *Retype\snew\s*\spassword:* %n\n *password\supdated\ssuccessfully* .
   #pam password change = yes
   map to guest = bad user
   # connection inactivity timeout in minutes
   deadtime = 5
   # enable registry based shares
   registry shares = yes
   # Make sure Samba isn't available over the Internet.
   # https://en.wikipedia.org/wiki/localhost
   # https://en.wikipedia.org/wiki/Private_network
   # https://en.wikipedia.org/wiki/Link-local_address
   # https://en.wikipedia.org/wiki/Unique_local_address
   hosts allow = 127.0.0.0/8 10.0.0.0/8 172.16.0.0/12 192.168.0.0/16 169.254.0.0/16 ::1
   hosts deny = all
'''  # noqa: E501


def _close_share(share_name):
    """Disconnect all samba users who are connected to the share."""
    subprocess.check_call(['smbcontrol', 'smbd', 'close-share', share_name])


def _conf_command(parameters, **kwargs):
    """Run samba configuration registry command."""
    subprocess.check_call(['net', 'conf'] + parameters, **kwargs)


def _create_share(mount_point, share_type, windows_filesystem=False):
    """Create samba public, group and private shares."""
    if share_type == 'open':
        subdir = 'open_share'
    elif share_type == 'group':
        subdir = 'group_share'
    elif share_type == 'home':
        subdir = 'homes'

    shares_path = _get_shares_path(mount_point)
    share_path = os.path.join(mount_point, shares_path, subdir)
    os.makedirs(share_path, exist_ok=True)

    # FAT and NTFS partitions don't support setting permissions
    if not windows_filesystem:
        if share_type in ['open', 'group']:
            _set_share_permissions(share_path)
        else:
            shutil.chown(share_path, group='users')
            os.chmod(share_path, 0o0775)

    share_name = _create_share_name(mount_point)

    if share_type == 'open':
        _define_open_share(share_name, share_path, windows_filesystem)
    elif share_type == 'group':
        _define_group_share(share_name + '_group', share_path,
                            windows_filesystem)
    elif share_type == 'home':
        _define_homes_share(share_name + '_home', share_path)


def _create_share_name(mount_point):
    """Create a share name."""
    share_name = os.path.basename(mount_point)
    if not share_name:
        share_name = 'disk'

    return share_name


def _define_open_share(name, path, windows_filesystem=False):
    """Define an open samba share."""
    try:
        _conf_command(['delshare', name], stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        pass
    _conf_command(['addshare', name, path, 'writeable=y', 'guest_ok=y'])
    if not windows_filesystem:
        _conf_command(['setparm', name, 'force group', 'freedombox-share'])
        _conf_command(['setparm', name, 'inherit permissions', 'yes'])


def _define_group_share(name, path, windows_filesystem=False):
    """Define a group samba share."""
    try:
        _conf_command(['delshare', name], stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        pass
    _conf_command(['addshare', name, path, 'writeable=y', 'guest_ok=n'])
    _conf_command(['setparm', name, 'valid users', '@freedombox-share @admin'])
    if not windows_filesystem:
        _conf_command(['setparm', name, 'force group', 'freedombox-share'])
        _conf_command(['setparm', name, 'inherit permissions', 'yes'])


def _define_homes_share(name, path):
    """Define a samba share for private homes."""
    try:
        _conf_command(['delshare', name], stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        pass
    userpath = os.path.join(path, '%u')
    _conf_command(['addshare', name, userpath, 'writeable=y', 'guest_ok=n'])
    _conf_command(['setparm', name, 'valid users', '@freedombox-share @admin'])
    _conf_command(
        ['setparm', name, 'preexec', 'mkdir -p -m 755 {}'.format(userpath)])


def _get_mount_point(path):
    """Get the mount point where the share is."""
    subpath = 'FreedomBox/shares/'
    if '/var/lib/freedombox/shares/' in path:
        if os.path.ismount(path.split('lib/freedombox/shares/')[0]):
            subpath = 'lib/freedombox/shares/'
        else:
            subpath = 'var/lib/freedombox/shares/'

    return path.split(subpath)[0]


def _get_shares() -> list[dict[str, str]]:
    """Get shares."""
    shares = []
    output = subprocess.check_output(['net', 'conf', 'list'])
    config = configparser.RawConfigParser()
    config.read_string(output.decode())
    for name in config.sections():
        share_type = 'open'
        if name.endswith('_group'):
            share_type = 'group'
        elif name.endswith('_home'):
            share_type = 'home'
        share_path = config[name]['path']
        mount_point = _get_mount_point(share_path)
        mount_point = os.path.normpath(mount_point)
        shares.append(
            dict(name=name, mount_point=mount_point, path=share_path,
                 share_type=share_type))

    return shares


def _get_shares_path(mount_point):
    """Return base path of the shared directories."""
    if mount_point == '/var':
        return 'lib/freedombox/shares/'
    var_directory = os.path.join(mount_point, 'var')

    if os.path.exists(var_directory) and os.stat(
            mount_point).st_dev == os.stat(var_directory).st_dev:
        return 'var/lib/freedombox/shares/'

    return 'FreedomBox/shares/'


def _set_open_share_permissions(directory):
    """Set file and directory permissions for open share."""
    shutil.chown(directory, group='freedombox-share')
    os.chmod(directory, 0o2775)
    for root, dirs, files in os.walk(directory):
        for subdir in dirs:
            subdir_path = os.path.join(root, subdir)
            shutil.chown(subdir_path, group='freedombox-share')
            os.chmod(subdir_path, 0o2775)
        for file in files:
            file_path = os.path.join(root, file)
            shutil.chown(file_path, group='freedombox-share')
            os.chmod(file_path, 0o0664)
    subprocess.check_call(['setfacl', '-Rm', 'g::rwX', directory])
    subprocess.check_call(['setfacl', '-Rdm', 'g::rwX', directory])


def _use_config_file(conf_file):
    """Set samba configuration file location."""
    import augeas
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.set('/augeas/load/Shellvars/lens', 'Shellvars.lns')
    aug.set('/augeas/load/Shellvars/incl[last() + 1]', DEFAULT_FILE)
    aug.load()

    aug.set('/files' + DEFAULT_FILE + '/SMBDOPTIONS',
            '--configfile={0}'.format(conf_file))
    aug.save()


def _set_share_permissions(directory):
    """Set file and directory permissions for a share."""
    shutil.chown(directory, group='freedombox-share')
    os.chmod(directory, 0o2775)
    for root, dirs, files in os.walk(directory):
        for subdir in dirs:
            subdir_path = os.path.join(root, subdir)
            shutil.chown(subdir_path, group='freedombox-share')
            os.chmod(subdir_path, 0o2775)
        for file in files:
            file_path = os.path.join(root, file)
            shutil.chown(file_path, group='freedombox-share')
            os.chmod(file_path, 0o0664)
    subprocess.check_call(['setfacl', '-Rm', 'g::rwX', directory])
    subprocess.check_call(['setfacl', '-Rdm', 'g::rwX', directory])


@privileged
def add_share(mount_point: str, share_type: str, windows_filesystem: bool):
    """Create a samba share."""
    if share_type not in ('open', 'group', 'home'):
        raise ValueError('Invalid share type')

    mount_point = os.path.normpath(mount_point)
    if not os.path.ismount(mount_point):
        raise RuntimeError(
            'Path "{0}" is not a mount point.'.format(mount_point))
    _create_share(mount_point, share_type, windows_filesystem)


@privileged
def delete_share(mount_point: str, share_type: str):
    """Delete a samba share configuration."""
    if share_type not in ('open', 'group', 'home'):
        raise ValueError('Invalid share type')

    mount_point = os.path.normpath(mount_point)
    shares = _get_shares()
    for share in shares:
        if share['mount_point'] == mount_point and share[
                'share_type'] == share_type:
            _close_share(share['name'])
            _conf_command(['delshare', share['name']])


@privileged
def get_shares() -> list[dict[str, str]]:
    """Get samba shares."""
    return _get_shares()


@privileged
def get_users() -> list[str]:
    """Get users from Samba database."""
    output = subprocess.check_output(['pdbedit', '-L']).decode()
    samba_users = [line.split(':')[0] for line in output.split()]
    return samba_users


@privileged
def setup():
    """Configure samba, use custom samba config file."""
    from plinth import action_utils
    with open(CONF_PATH, 'w', encoding='utf-8') as file_handle:
        file_handle.write(CONF)
    _use_config_file(CONF_PATH)
    os.makedirs('/var/lib/freedombox', exist_ok=True)
    os.chmod('/var/lib/freedombox', 0o0755)

    # Disable NetBIOS Service, used with now deprecated SMB1 protocol
    if action_utils.service_is_running('nmbd'):
        action_utils.service_stop('nmbd')
    action_utils.service_disable('nmbd')
    action_utils.service_mask('nmbd')

    if action_utils.service_is_running('smbd'):
        action_utils.service_restart('smbd')


@privileged
def dump_shares():
    """Dump registy share configuration."""
    os.makedirs(os.path.dirname(SHARES_CONF_BACKUP_FILE), exist_ok=True)
    with open(SHARES_CONF_BACKUP_FILE, 'w', encoding='utf-8') as backup_file:
        command = ['net', 'conf', 'list']
        subprocess.run(command, stdout=backup_file, check=True)


@privileged
def restore_shares():
    """Restore registy share configuration."""
    if not os.path.exists(SHARES_CONF_BACKUP_FILE):
        raise RuntimeError(
            'Backup file {0} does not exist.'.format(SHARES_CONF_BACKUP_FILE))
    _conf_command(['drop'])
    _conf_command(['import', SHARES_CONF_BACKUP_FILE])


@privileged
def uninstall():
    """Drop all Samba shares."""
    _conf_command(['drop'])
    for path in [CONF_PATH, DEFAULT_FILE]:
        # Both files are typically created on setup()
        pathlib.Path(path).unlink(missing_ok=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Setup CherryPy web server.
"""

import logging
import os
import sys
import warnings
from typing import ClassVar

import cherrypy

from . import app as app_module
from . import cfg, log, web_framework

logger = logging.getLogger(__name__)


def _mount_static_directory(static_dir, static_url):
    config = {
        '/': {
            'tools.staticdir.root': static_dir,
            'tools.staticdir.on': True,
            'tools.staticdir.dir': '.'
        }
    }
    app = cherrypy.tree.mount(None, static_url, config)
    log.setup_cherrypy_static_directory(app)


def init():
    """Setup CherryPy server"""
    logger.info('Setting up CherryPy server')

    # Configure default server
    cherrypy.config.update({
        'server.max_request_body_size': 0,
        'server.socket_host': cfg.host,
        'server.socket_port': cfg.port,
        'server.thread_pool': 10,
        # Avoid stating files once per second in production
        'engine.autoreload.on': cfg.develop,
    })

    application = web_framework.get_wsgi_application()
    cherrypy.tree.graft(application, cfg.server_dir)

    static_dir = os.path.join(cfg.file_root, 'static')
    _mount_static_directory(static_dir, web_framework.get_static_url())

    custom_static_dir = cfg.custom_static_dir
    custom_static_url = '/plinth/custom/static'
    if os.path.exists(custom_static_dir):
        _mount_static_directory(custom_static_dir, custom_static_url)
    else:
        logger.debug(
            'Not serving custom static directory %s on %s, '
            'directory does not exist', custom_static_dir, custom_static_url)

    _mount_static_directory('/usr/share/javascript', '/javascript')

    for app in app_module.App.list():
        module = sys.modules[app.__module__]
        module_path = os.path.dirname(module.__file__)
        static_dir = os.path.join(module_path, 'static')
        if not os.path.isdir(static_dir):
            continue

        urlprefix = "%s%s" % (web_framework.get_static_url(), app.app_id)
        _mount_static_directory(static_dir, urlprefix)

    for component in StaticFiles.list():
        component.mount()

    cherrypy.engine.signal_handler.subscribe()


def run(on_web_server_stop):
    """Start the web server and block it until exit."""
    with warnings.catch_warnings():
        # Suppress warning that some of the static directories don't exist.
        # Since there is no way to add/remove those tree mounts at will, we
        # need to add them all before hand even if they don't exist now. If the
        # directories becomes available later, CherryPy serves them just fine.
        warnings.filterwarnings(
            'ignore', '(.|\n)*is not an existing filesystem path(.|\n)*',
            UserWarning)
        cherrypy.engine.start()

    cherrypy.engine.subscribe('stop', on_web_server_stop)
    cherrypy.engine.block()


class StaticFiles(app_module.FollowerComponent):
    """Component to serve static files shipped with an app.

    Any files in <app>/static directory will be automatically served on
    /static/<app>/ directory by FreedomBox. This allows each app to ship custom
    static files that are served by the web server.

    However, in some rare circumstances, a system folder will need to be served
    on a path for the app to work. This component allows declaring such
    directories and the web paths they should be served on.

    """

    _all_instances: ClassVar[dict[str, 'StaticFiles']] = {}

    def __init__(self, component_id, directory_map=None):
        """Initialize the component.

        component_id should be a unique ID across all components of an app and
        across all components.

        directory_map should be a dictionary with keys to be web paths and
        values to be absolute path of the directory on disk to serve. The
        static files from the directory are served over the given web path. The
        web path will be prepended with the FreedomBox's configured base web
        path. For example, {'/foo': '/usr/share/foo'} means that
        '/usr/share/foo/bar.png' will be served over '/plinth/foo/bar.png' if
        FreedomBox is configured to be served on '/plinth'.

        """
        super().__init__(component_id)
        self.directory_map = directory_map
        self._all_instances[component_id] = self

    @classmethod
    def list(cls):
        """Return a list of all instances."""
        return cls._all_instances.values()

    def mount(self):
        """Perform configuration of the web server to handle static files.

        Called by web server abstraction layer after web server has been setup.

        """
        if self.directory_map:
            for web_path, file_path in self.directory_map.items():
                web_path = '%s%s' % (cfg.server_dir, web_path)
                _mount_static_directory(file_path, web_path)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manfiest for letsencrypt.
"""

# XXX: Backup and restore the Apache site configuration.
backup = {'secrets': {'directories': ['/etc/letsencrypt/']}}



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for Tor Proxy app.
"""

import pytest

from plinth.tests import functional

_TOR_FEATURE_TO_ELEMENT = {'software': 'torproxy-apt_transport_tor_enabled'}

pytestmark = [pytest.mark.apps, pytest.mark.torproxy]


class TestTorProxyApp(functional.BaseAppTests):
    """Tests for the Tor Proxy app."""
    app_name = 'torproxy'
    has_service = True
    has_web = False
    # TODO: Investigate why accessing IPv6 sites through Tor fails in
    # container.
    check_diagnostics = False

    def test_set_download_software_packages_over_tor(self, session_browser):
        """Test setting download software packages over Tor."""
        functional.app_enable(session_browser, 'torproxy')
        _feature_enable(session_browser, 'software', should_enable=True)
        _feature_enable(session_browser, 'software', should_enable=False)
        _assert_feature_enabled(session_browser, 'software', enabled=False)

    @pytest.mark.backups
    def test_backup_restore(self, session_browser):
        """Test backup and restore of configuration."""
        functional.app_enable(session_browser, 'torproxy')
        # TODO: Check that upstream bridges are restored.
        functional.backup_create(session_browser, 'torproxy', 'test_torproxy')

        functional.backup_restore(session_browser, 'torproxy', 'test_torproxy')

        assert functional.service_is_running(session_browser, 'torproxy')


def _feature_enable(browser, feature, should_enable):
    """Enable/disable a Tor Proxy feature."""
    element_name = _TOR_FEATURE_TO_ELEMENT[feature]
    functional.nav_to_module(browser, 'torproxy')
    checkbox_element = browser.find_by_name(element_name).first
    if should_enable == checkbox_element.checked:
        return

    if should_enable:
        checkbox_element.check()
    else:
        checkbox_element.uncheck()

    functional.submit(browser, form_class='form-configuration')
    functional.wait_for_config_update(browser, 'torproxy')


def _assert_feature_enabled(browser, feature, enabled):
    """Assert whether Tor Proxy feature is enabled or disabled."""
    element_name = _TOR_FEATURE_TO_ELEMENT[feature]
    functional.nav_to_module(browser, 'torproxy')
    assert browser.find_by_name(element_name).first.checked == enabled



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app for configuring ikiwiki."""

from django.contrib import messages
from django.shortcuts import redirect
from django.template.response import TemplateResponse
from django.urls import reverse_lazy
from django.utils.translation import gettext as _

from plinth import app as app_module
from plinth import views

from . import privileged
from .forms import IkiwikiCreateForm


class IkiwikiAppView(views.AppView):
    """Serve configuration page."""

    app_id = 'ikiwiki'
    template_name = 'ikiwiki_configure.html'

    def get_context_data(self, **kwargs):
        """Return the context data for rendering the template view."""
        app = app_module.App.get('ikiwiki')
        sites = app.refresh_sites()
        sites = [name for name in sites if name != '']

        context = super().get_context_data(**kwargs)
        context['sites'] = sites
        return context


def create(request):
    """Form to create a wiki or blog."""
    form = None

    app = app_module.App.get('ikiwiki')
    if request.method == 'POST':
        form = IkiwikiCreateForm(request.POST, prefix='ikiwiki')
        if form.is_valid():
            if form.cleaned_data['site_type'] == 'wiki':
                _create_wiki(request, form.cleaned_data['name'],
                             form.cleaned_data['admin_name'],
                             form.cleaned_data['admin_password'])
            elif form.cleaned_data['site_type'] == 'blog':
                _create_blog(request, form.cleaned_data['name'],
                             form.cleaned_data['admin_name'],
                             form.cleaned_data['admin_password'])

            app.refresh_sites()
            if app.is_enabled():
                app.set_enabled(True)

            return redirect(reverse_lazy('ikiwiki:index'))
    else:
        form = IkiwikiCreateForm(prefix='ikiwiki')

    return TemplateResponse(request, 'ikiwiki_create.html', {
        'title': app.info.name,
        'form': form
    })


def _create_wiki(request, name, admin_name, admin_password):
    """Create wiki."""
    try:
        privileged.create_wiki(name, admin_name, admin_password)
        messages.success(request, _('Created wiki {name}.').format(name=name))
    except Exception as error:
        messages.error(request,
                       _('Could not create wiki: {error}').format(error=error))


def _create_blog(request, name, admin_name, admin_password):
    """Create blog."""
    try:
        privileged.create_blog(name, admin_name, admin_password)
        messages.success(request, _('Created blog {name}.').format(name=name))
    except Exception as error:
        messages.error(request,
                       _('Could not create blog: {error}').format(error=error))


def delete(request, name):
    """Handle deleting wikis/blogs, showing a confirmation dialog first.

    On GET, display a confirmation page.
    On POST, delete the wiki/blog.
    """
    app = app_module.App.get('ikiwiki')
    title = app.components['shortcut-ikiwiki-' + name].name
    if request.method == 'POST':
        try:
            privileged.delete(name)
            app.remove_shortcut(name)
            messages.success(request,
                             _('{title} deleted.').format(title=title))
        except Exception as error:
            messages.error(
                request,
                _('Could not delete {title}: {error}').format(
                    title=title, error=error))

        return redirect(reverse_lazy('ikiwiki:index'))

    return TemplateResponse(request, 'ikiwiki_delete.html', {
        'title': app.info.name,
        'name': title
    })



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app for configuring WordPress."""

from django.contrib import messages
from django.utils.translation import gettext as _

from plinth import views

from . import privileged
from .forms import WordPressForm


class WordPressAppView(views.AppView):
    """Serve configuration page."""

    form_class = WordPressForm
    app_id = 'wordpress'

    def get_initial(self):
        """Get the current WordPress settings."""
        status = super().get_initial()
        status['is_public'] = privileged.is_public()
        return status

    def form_valid(self, form):
        """Apply the changes submitted in the form."""
        old_status = form.initial
        new_status = form.cleaned_data
        if old_status['is_public'] != new_status['is_public']:
            privileged.set_public(new_status['is_public'])
            messages.success(self.request, _('Configuration updated'))

        return super().form_valid(form)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
App component to manage users and groups.
"""

import itertools
from typing import ClassVar

from plinth import app


class UsersAndGroups(app.FollowerComponent):
    """Component to manage users and groups of an app."""

    # Class variable to hold a list of user groups for apps
    _all_components: ClassVar[set['UsersAndGroups']] = set()

    def __init__(self, component_id, reserved_usernames=[], groups={}):
        """Store reserved_usernames and groups of the app.

        'reserved_usernames' is a list of operating system user names that the
        app uses. It is not permitted to create a FreedomBox user with one of
        these names.

        'groups' is a dictionary of the following format: {"group_name": "A
        localized string describing what permissions are offered to the users
        of this group"}.

        """
        super().__init__(component_id)

        self.reserved_usernames = reserved_usernames
        self.groups = groups

        self._all_components.add(self)

    @classmethod
    def get_groups(cls):
        """Return a set of all groups."""
        all_groups = itertools.chain(*(component.groups.keys()
                                       for component in cls._all_components))
        return set(all_groups)

    @classmethod
    def get_group_choices(cls):
        """Return list of groups that can be used as form choices."""
        all_groups = itertools.chain(*(component.groups.items()
                                       for component in cls._all_components))
        choices = [(group, f'{description} ({group})')
                   for group, description in set(all_groups)]
        return sorted(choices, key=lambda g: g[0])

    @classmethod
    def is_username_reserved(cls, username):
        """Returns whether the given username is reserved or not."""
        return any((username in component.reserved_usernames
                    for component in cls._all_components))



# SPDX-License-Identifier: AGPL-3.0-or-later
"""App component for other apps to use backup/restore functionality."""

import copy

from plinth import app

from . import privileged


def _validate_directories_and_files(section):
    """Validate directories and files keys in a section."""
    if not section:
        return

    assert isinstance(section, dict)

    if 'directories' in section:
        assert isinstance(section['directories'], list)
        for directory in section['directories']:
            assert isinstance(directory, str)

    if 'files' in section:
        assert isinstance(section['files'], list)
        for file_path in section['files']:
            assert isinstance(file_path, str)


def _validate_services(services):
    """Validate services manifest provided as list."""
    if not services:
        return

    assert isinstance(services, list)
    for service in services:
        assert isinstance(service, (str, dict))
        if isinstance(service, dict):
            _validate_service(service)


def _validate_service(service):
    """Validate a service manifest provided as a dictionary."""
    assert isinstance(service['name'], str)
    assert isinstance(service['type'], str)
    assert service['type'] in ('apache', 'uwsgi', 'system')
    if service['type'] == 'apache':
        assert service['kind'] in ('config', 'site', 'module')


def _validate_settings(settings):
    """Validate settings stored by an in kvstore."""
    if not settings:
        return

    assert isinstance(settings, list)
    for setting in settings:
        assert isinstance(setting, str)


class BackupRestore(app.FollowerComponent):
    """Component to backup/restore an app."""

    def __init__(self, component_id, config=None, data=None, secrets=None,
                 services=None, settings=None):
        """Initialize the backup/restore component."""
        super().__init__(component_id)

        _validate_directories_and_files(config)
        self.config = config or {}
        _validate_directories_and_files(data)
        self._data = data or {}
        _validate_directories_and_files(secrets)
        self.secrets = secrets or {}
        _validate_services(services)
        self.services = services or []
        _validate_settings(settings)
        self.settings = settings or []

        self.has_data = (bool(config) or bool(data) or bool(secrets)
                         or bool(settings))

    def __eq__(self, other):
        """Check if this component is same as another."""
        return self.component_id == other.component_id

    @property
    def data(self):
        """Add additional files to data files list."""
        data = copy.deepcopy(self._data)
        settings_file = self._get_settings_file()
        if settings_file:
            data.setdefault('files', []).append(settings_file)

        return data

    @property
    def manifest(self):
        """Return the backup details as a dictionary."""
        manifest = {}
        if self.config:
            manifest['config'] = self.config

        if self.secrets:
            manifest['secrets'] = self.secrets

        if self.data:
            manifest['data'] = self.data

        if self.services:
            manifest['services'] = self.services

        if self.settings:
            manifest['settings'] = self.settings

        return manifest

    def backup_pre(self, packet):
        """Perform any special operations before backup."""
        self._settings_backup_pre()

    def backup_post(self, packet):
        """Perform any special operations after backup."""

    def restore_pre(self, packet):
        """Perform any special operations before restore."""

    def restore_post(self, packet):
        """Perform any special operations after restore."""
        self._settings_restore_post()

    def _get_settings_file(self):
        """Return the settings file path to list of files to backup."""
        if not self.settings or not self.app_id:
            return None

        data_path = '/var/lib/plinth/backups-data/'
        return data_path + f'{self.app_id}-settings.json'

    def _settings_backup_pre(self):
        """Read keys from kvstore and store them in a file to backup."""
        if not self.settings:
            return

        from plinth import kvstore
        data = {}
        for key in self.settings:
            try:
                data[key] = kvstore.get(key)
            except Exception:
                pass

        privileged.dump_settings(self.app_id, data)

    def _settings_restore_post(self):
        """Read from a file and restore keys to kvstore."""
        if not self.settings:
            return

        data = privileged.load_settings(self.app_id)

        from plinth import kvstore
        for key, value in data.items():
            kvstore.set(key, value)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Privileged operations for managing aliases."""

import pathlib
import shutil

from plinth.actions import privileged


@privileged
def setup_aliases():
    """Create a the sqlite3 database to be managed by FreedomBox."""
    path = pathlib.Path('/var/lib/postfix/freedombox-aliases/')
    path.mkdir(mode=0o750, exist_ok=True)
    shutil.chown(path, user='plinth', group='postfix')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the calibre module.
"""

from django.urls import re_path

from . import views

urlpatterns = [
    re_path(r'^apps/calibre/$', views.CalibreAppView.as_view(), name='index'),
    re_path(r'^apps/calibre/library/create/$',
            views.CreateLibraryView.as_view(), name='create-library'),
    re_path(r'^apps/calibre/library/(?P<name>[a-zA-Z0-9_.-]+)/delete/$',
            views.delete_library, name='delete-library'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configuration helper for WordPress."""

import os
import pathlib
import random
import shutil
import string
import subprocess

import augeas

from plinth import action_utils
from plinth.actions import privileged

_public_access_file = pathlib.Path('/etc/wordpress/is_public')
_config_file_path = pathlib.Path('/etc/wordpress/config-default.php')
_static_config_file_path = pathlib.Path('/etc/wordpress/freedombox-static.php')
_db_file_path = pathlib.Path('/etc/wordpress/database.php')
_db_backup_file = pathlib.Path(
    '/var/lib/plinth/backups-data/wordpress-database.sql')
DB_HOST = 'localhost'
DB_NAME = 'wordpress_fbx'
DB_USER = 'wordpress_fbx'


@privileged
def setup():
    """Create initial configuration and database for WordPress."""
    if _db_file_path.exists() or _config_file_path.exists():
        if _config_file_path.exists():
            _upgrade_config_file()

        return

    db_password = _generate_secret_key(16)

    _create_config_file(DB_HOST, DB_NAME, DB_USER, db_password)
    with action_utils.service_ensure_running('mysql'):
        _create_database(DB_NAME)
        _set_privileges(DB_HOST, DB_NAME, DB_USER, db_password)


def _create_config_file(db_host, db_name, db_user, db_password):
    """Create a PHP configuration file included by WordPress."""
    secret_keys = [_generate_secret_key() for _ in range(8)]

    config_contents = f'''<?php
# Created by FreedomBox
include_once('{_static_config_file_path}');
include_once('{_db_file_path}');
define('DB_NAME', $dbname);
define('DB_USER', $dbuser);
define('DB_PASSWORD', $dbpass);
define('DB_HOST', $dbserver);

define('AUTH_KEY', '{secret_keys[0]}');
define('SECURE_AUTH_KEY', '{secret_keys[1]}');
define('LOGGED_IN_KEY', '{secret_keys[2]}');
define('NONCE_KEY', '{secret_keys[3]}');
define('AUTH_SALT', '{secret_keys[4]}');
define('SECURE_AUTH_SALT', '{secret_keys[5]}');
define('LOGGED_IN_SALT', '{secret_keys[6]}');
define('NONCE_SALT', '{secret_keys[7]}');

define('WP_CONTENT_DIR', '/var/lib/wordpress/wp-content');

define('DISABLE_WP_CRON', true);
'''
    _config_file_path.write_text(config_contents, encoding='utf-8')

    db_contents = f'''<?php
# Created by FreedomBox
$dbuser='{db_user}';
$dbpass='{db_password}';
$dbname='{db_name}';
$dbserver='{db_host}';
'''
    old_umask = os.umask(0o037)
    try:
        _db_file_path.write_text(db_contents, encoding='utf-8')
    finally:
        os.umask(old_umask)

    shutil.chown(_db_file_path, group='www-data')


def _create_database(db_name):
    """Create an empty MySQL database for WordPress."""
    # Wordpress' install.php creates the tables.
    # SQL injection is avoided due to known input.
    query = f'''CREATE DATABASE {db_name};'''
    subprocess.run(['mysql', '--user', 'root'], input=query.encode(),
                   check=True)


def _set_privileges(db_host, db_name, db_user, db_password):
    """Create user, set password and provide permissions on the database."""
    # SQL injection is avoided due to known input.
    query = f'''GRANT SELECT,INSERT,UPDATE,DELETE,CREATE,DROP,ALTER
  ON {db_name}.*
  TO {db_user}@{db_host}
  IDENTIFIED BY '{db_password}';
FLUSH PRIVILEGES;
'''
    subprocess.run(['mysql', '--user', 'root'], input=query.encode(),
                   check=True)


def _generate_secret_key(length=64, chars=None):
    """Generate a new random secret key for use with WordPress."""
    chars = chars or (string.ascii_letters + string.digits +
                      '!@#$%^&*()-_ []{}<>~`+=,.:/?|')
    rand = random.SystemRandom()
    return ''.join(rand.choice(chars) for _ in range(length))


def _upgrade_config_file():
    """Upgrade existing config file to add changes."""
    include_line = f"include_once('{_static_config_file_path}');"
    lines = _config_file_path.read_text(encoding='utf-8').splitlines()
    if include_line not in lines:
        lines.insert(2, include_line)  # Insert on 3rd line
        _config_file_path.write_text('\n'.join(lines), encoding='utf-8')


@privileged
def set_public(enable: bool):
    """Allow/disallow public access."""
    if enable:
        _public_access_file.touch()
    else:
        _public_access_file.unlink(missing_ok=True)

    action_utils.service_reload('apache2')


def is_public() -> bool:
    """Return whether public access is enabled."""
    return _public_access_file.exists()


@privileged
def dump_database():
    """Dump database to file."""
    _db_backup_file.parent.mkdir(parents=True, exist_ok=True)
    with action_utils.service_ensure_running('mysql'):
        with _db_backup_file.open('w', encoding='utf-8') as file_handle:
            subprocess.run([
                'mysqldump', '--add-drop-database', '--add-drop-table',
                '--add-drop-trigger', '--user', 'root', '--databases', DB_NAME
            ], stdout=file_handle, check=True)


@privileged
def restore_database():
    """Restore database from file."""
    with action_utils.service_ensure_running('mysql'):
        with _db_backup_file.open('r', encoding='utf-8') as file_handle:
            subprocess.run(['mysql', '--user', 'root'], stdin=file_handle,
                           check=True)

        _set_privileges(DB_HOST, DB_NAME, DB_USER, _read_db_password())


def _read_db_password():
    """Return the password stored in the DB configuration file."""
    aug = _load_augeas()
    return aug.get('./$dbpass').strip('\'"')


def _load_augeas():
    """Initialize augeas."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.transform('Phpvars', str(_db_file_path))
    aug.set('/augeas/context', '/files' + str(_db_file_path))
    aug.load()

    return aug


@privileged
def uninstall():
    """Remove config files and drop database."""
    _drop_database(DB_HOST, DB_NAME, DB_USER)
    for file_ in [_public_access_file, _config_file_path, _db_file_path]:
        file_.unlink(missing_ok=True)


def _drop_database(db_host, db_name, db_user):
    """Drop the mysql database that was created during install."""
    with action_utils.service_ensure_running('mysql'):
        query = f"DROP DATABASE {db_name};"
        subprocess.run(['mysql', '--user', 'root'], input=query.encode(),
                       check=False)

        query = f"DROP USER IF EXISTS {db_user}@{db_host};"
        subprocess.run(['mysql', '--user', 'root'], input=query.encode(),
                       check=False)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the Gitweb module.
"""

from django.urls import re_path

from .views import CreateRepoView, EditRepoView, GitwebAppView, delete

urlpatterns = [
    re_path(r'^apps/gitweb/$', GitwebAppView.as_view(), name='index'),
    re_path(r'^apps/gitweb/create/$', CreateRepoView.as_view(), name='create'),
    re_path(
        r'^apps/gitweb/(?P<name>[a-zA-Z0-9-._]+)/edit/$',
        EditRepoView.as_view(),
        name='edit',
    ),
    re_path(
        r'^apps/gitweb/(?P<name>[a-zA-Z0-9-._]+)/delete/$',
        delete,
        name='delete',
    ),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Privileged actions for Kiwix content server.
"""

import os
import pathlib
import shutil
import subprocess
from xml.etree import ElementTree

from plinth import action_utils
from plinth.actions import privileged
from plinth.modules import kiwix

# Only one central library is supported.
KIWIX_HOME = pathlib.Path('/var/lib/kiwix-server-freedombox')
LIBRARY_FILE = KIWIX_HOME / 'library_zim.xml'
CONTENT_DIR = KIWIX_HOME / 'content'


@privileged
def add_package(file_name: str):
    """Adds a content package to Kiwix.

    Adding packages is idempotent.

    Users can add content to Kiwix in multiple ways:
     - Upload a ZIM file
     - Provide a link to the ZIM file
     - Provide a magnet link to the ZIM file

    The commandline download manager aria2c is a dependency of kiwix-tools.
    aria2c is used for both HTTP and Magnet downloads.
    """
    kiwix.validate_file_name(file_name)

    # Moving files to the Kiwix library path ensures that
    # they can't be removed by other apps or users.
    zim_file_name = pathlib.Path(file_name).name
    CONTENT_DIR.mkdir(exist_ok=True)
    zim_file_dest = str(CONTENT_DIR / zim_file_name)
    shutil.chown(file_name, 'root', 'root')
    os.chmod(file_name, 0o644)
    shutil.move(file_name, zim_file_dest)

    _kiwix_manage_add(zim_file_dest)


def _kiwix_manage_add(zim_file: str):
    subprocess.check_call(['kiwix-manage', LIBRARY_FILE, 'add', zim_file])

    # kiwix-serve doesn't read the library file unless it is restarted.
    action_utils.service_try_restart('kiwix-server-freedombox')


@privileged
def uninstall() -> None:
    """Remove all content during uninstall."""
    shutil.rmtree(str(CONTENT_DIR), ignore_errors=True)
    LIBRARY_FILE.unlink(missing_ok=True)


@privileged
def list_packages() -> dict[str, dict[str, str]]:
    """Return the list of content packages configured in library file."""
    library = ElementTree.parse(LIBRARY_FILE).getroot()

    books = {}
    for book in library:
        path = book.attrib['path'].split('/')[-1]
        path = path.removesuffix('.zim').lower()  # Strip '.zim' from the path
        try:
            books[book.attrib['id']] = {
                'title': book.attrib['title'],
                'description': book.attrib['description'],
                'path': path
            }
        except KeyError:
            pass  # Ignore entries that don't have expected properties

    return books


@privileged
def delete_package(zim_id: str):
    """Remove a content package from the library file."""
    library = ElementTree.parse(LIBRARY_FILE).getroot()

    for book in library:
        try:
            if book.attrib['id'] != zim_id:
                continue

            subprocess.check_call(
                ['kiwix-manage', LIBRARY_FILE, 'remove', zim_id])
            (KIWIX_HOME / book.attrib['path']).unlink()
            action_utils.service_try_restart('kiwix-server-freedombox')
            return
        except KeyError:  # Expected properties not found on elements
            pass



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app for configuring MediaWiki.
"""

import json
import logging
import pathlib
import subprocess

from django import forms
from django.core import validators
from django.utils.translation import gettext_lazy as _

from . import privileged


def get_skins():
    """Return a list of available skins as choice field values."""
    skins_dir = pathlib.Path('/var/lib/mediawiki/skins')
    if not skins_dir.exists():
        return []

    return [(skin.name.lower(), skin.name) for skin in skins_dir.iterdir()
            if skin.is_dir()]


def get_languages():
    """Return a list of available languages as choice field values."""
    # Names.php has different locations on Bullseye Bookworm
    names_old = pathlib.Path('/usr/share/mediawiki/languages/data/Names.php')
    names_new = pathlib.Path('/usr/share/mediawiki/includes/languages/'
                             'data/Names.php')
    names_file = names_old
    if not names_old.exists():
        names_file = names_new

    script = rf'''<?php
    require_once('{names_file}');
    print(json_encode(Mediawiki\Languages\Data\Names::$names));'''
    try:
        process = subprocess.run([privileged.get_php_command()],
                                 input=script.encode(), stdout=subprocess.PIPE,
                                 check=True)
        languages_dict = json.loads(process.stdout.decode())
        return list(languages_dict.items())
    except Exception as exception:
        logging.exception('Unable read list of Mediawiki languages.',
                          exception)
        return [('en', 'English')]


class MediaWikiForm(forms.Form):  # pylint: disable=W0232
    """MediaWiki configuration form."""
    password = forms.CharField(
        label=_('Administrator Password'), help_text=_(
            'Set a new password for MediaWiki\'s administrator account '
            '(admin). The password cannot be a common one and the minimum '
            'required length is <strong>10 characters</strong>. '
            'Leave this field blank to keep the current password.'),
        required=False, widget=forms.PasswordInput, min_length=10)

    domain = forms.CharField(
        label=_('Domain'), required=False, help_text=_(
            'Used by MediaWiki to generate URLs that point to the wiki '
            'such as in footer, feeds and emails. Examples: '
            '"myfreedombox.example.org" or "example.onion".'),
        validators=[validators.RegexValidator('[$"]', inverse_match=True)])

    site_name = forms.CharField(
        label=_('Site Name'), required=False,
        help_text=_('Name of the site as displayed throughout the wiki.'),
        validators=[validators.RegexValidator('[$"]', inverse_match=True)])

    enable_public_registrations = forms.BooleanField(
        label=_('Enable public registrations'), required=False,
        help_text=_('If enabled, anyone on the internet will be able to '
                    'create an account on your MediaWiki instance.'))

    enable_private_mode = forms.BooleanField(
        label=_('Enable private mode'), required=False,
        help_text=_('If enabled, access will be restricted. Only people '
                    'who have accounts can read/write to the wiki. '
                    'Public registrations will also be disabled.'))

    default_skin = forms.ChoiceField(
        label=_('Default Skin'), required=False,
        help_text=_('Choose a default skin for your MediaWiki installation. '
                    'Users have the option to select their preferred skin.'),
        choices=get_skins)

    default_lang = forms.ChoiceField(
        label=_('Default Language'), required=False,
        help_text=_('Choose a default language for your MediaWiki '
                    'installation. Users have the option to select '
                    'their preferred language.'), choices=get_languages)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Views for the JSXC module."""

from django.http import Http404
from django.views.generic import TemplateView

import plinth.app as app_module
from plinth.modules import config


class JsxcView(TemplateView):
    """A simple page to embed Javascript XMPP Client library."""

    template_name = 'jsxc_launch.html'

    def dispatch(self, request, *args, **kwargs):
        """Don't serve the view when app is disabled."""
        app = app_module.App.get('jsxc')
        if not app.is_enabled():
            raise Http404

        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, *args, **kwargs):
        """Add domain information to view context."""
        context = super().get_context_data(*args, **kwargs)
        context['domainname'] = config.get_domainname()
        return context



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Views for security module."""

from django.contrib import messages
from django.template.response import TemplateResponse
from django.utils.translation import gettext as _

from plinth import action_utils
from plinth.modules import security
from plinth.modules.upgrades import is_backports_requested
from plinth.privileged import service as service_privileged
from plinth.views import AppView

from .forms import SecurityForm


class SecurityAppView(AppView):
    """Show security app main page."""

    app_id = 'security'
    template_name = 'security.html'
    form_class = SecurityForm
    prefix = 'security'

    def get_initial(self):
        """Return the values to fill in the form."""
        initial = super().get_initial()
        initial.update(get_status(self.request))
        return initial

    def get_context_data(self, *args, **kwargs):
        """Add additional context data for template."""
        context = super().get_context_data(*args, **kwargs)
        context['is_backports_requested'] = is_backports_requested()
        return context

    def form_valid(self, form):
        """Apply the changes submitted in the form."""
        _apply_changes(self.request, form.initial, form.cleaned_data)
        return super().form_valid(form)


def get_status(request):
    """Return the current status."""
    return {'fail2ban_enabled': action_utils.service_is_enabled('fail2ban')}


def _apply_changes(request, old_status, new_status):
    """Apply the form changes."""
    if old_status['fail2ban_enabled'] != new_status['fail2ban_enabled']:
        if new_status['fail2ban_enabled']:
            service_privileged.enable('fail2ban')
        else:
            service_privileged.disable('fail2ban')

        messages.success(request, _('Configuration updated.'))


def report(request):
    """Serve the security report page."""
    apps_report = security.get_apps_report()
    return TemplateResponse(
        request, 'security_report.html', {
            'title':
                _('Security Report'),
            'freedombox_report':
                apps_report.pop('freedombox'),
            'apps_report':
                sorted(apps_report.values(), key=lambda app: app['name']),
        })



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for WordPress.
"""

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.wordpress]


@pytest.fixture(scope='module', autouse=True)
def fixture_background(session_browser):
    """Login and install the app."""
    functional.login(session_browser)
    functional.install(session_browser, 'wordpress')
    yield
    functional.app_disable(session_browser, 'wordpress')


def test_enable_disable(session_browser):
    """Test enabling the app."""
    functional.app_disable(session_browser, 'wordpress')

    functional.app_enable(session_browser, 'wordpress')
    assert functional.service_is_running(session_browser, 'wordpress')
    assert functional.is_available(session_browser, 'wordpress')

    functional.app_disable(session_browser, 'wordpress')
    assert functional.service_is_not_running(session_browser, 'wordpress')
    assert not functional.is_available(session_browser, 'wordpress')


def test_post(session_browser):
    """Test writing a blog post."""
    functional.app_enable(session_browser, 'wordpress')
    _write_post(session_browser, 'FunctionalTest')
    assert _get_post(session_browser, 'FunctionalTest')
    _delete_post(session_browser, 'FunctionalTest')
    assert not _get_post(session_browser, 'FunctionalTest')


def test_public_mode(session_browser):
    """Test that site is available without login in public mode."""
    functional.app_enable(session_browser, 'wordpress')
    _enable_public_mode(session_browser, True)

    def no_login_prompt():
        _load_site(session_browser)
        return not functional.is_login_prompt(session_browser)

    try:
        functional.logout(session_browser)
        functional.eventually(no_login_prompt)
    finally:
        functional.login(session_browser)


def test_private_mode(session_browser):
    """Test that site is not available without login in public mode."""
    functional.app_enable(session_browser, 'wordpress')
    _enable_public_mode(session_browser, False)

    def login_prompt():
        _load_site(session_browser)
        return functional.is_login_prompt(session_browser)

    try:
        functional.logout(session_browser)
        functional.eventually(login_prompt)
    finally:
        functional.login(session_browser)


@pytest.mark.backups
def test_backup(session_browser):
    """Test backing up and restoring."""
    functional.app_enable(session_browser, 'wordpress')
    _write_post(session_browser, 'FunctionalTest')
    functional.backup_create(session_browser, 'wordpress', 'test_wordpress')
    _delete_post(session_browser, 'FunctionalTest')
    functional.uninstall(session_browser, 'wordpress')
    functional.backup_restore(session_browser, 'wordpress', 'test_wordpress')
    assert _get_post(session_browser, 'FunctionalTest')


def _load_site(browser):
    """Visit WordPress site and wait until becomes available."""
    functional.visit(browser, '/wordpress/wp-admin/')

    def loaded():
        browser.reload()
        title_node = browser.find_by_css('title')
        return (not title_node or '404' not in title_node[0].text)

    functional.eventually(loaded)


def _visit_site(browser):
    """Visit WordPress and run the first setup wizard if needed."""
    _load_site(browser)
    if '/install.php' in browser.url:
        # continue past language selection screen
        if browser.find_by_id('language-continue'):
            browser.find_by_id('language-continue').click()

        browser.fill('weblog_title', 'Test Blog')
        browser.fill('user_name', functional.config['DEFAULT']['username'])
        # browser.fill() once does not work for some reason for password field
        browser.fill('admin_password',
                     functional.config['DEFAULT']['password'])
        browser.fill('admin_password',
                     functional.config['DEFAULT']['password'])
        browser.check('pw_weak')
        browser.fill('admin_email', 'admin@example.org')
        submit_button = browser.find_by_id('submit')
        functional.submit(browser, element=submit_button)

        if not browser.find_by_css('.install-success'):
            raise Exception('WordPress installation failed')

        functional.visit(browser, '/wordpress/wp-admin/')

    if not browser.find_by_id('wpadminbar'):
        functional.visit(browser, '/wordpress/wp-login.php')
        browser.fill('log', functional.config['DEFAULT']['username'])
        browser.fill('pwd', functional.config['DEFAULT']['password'])
        login_button = browser.find_by_id('wp-submit')
        functional.submit(browser, element=login_button)


def _write_post(browser, title):
    """Create a blog post in WordPress site."""
    post = _get_post(browser, title)
    if post:
        _delete_post(browser, title)

    functional.visit(browser, '/wordpress/wp-admin/post-new.php')
    if browser.find_by_css('.edit-post-welcome-guide'):
        browser.find_by_css('.components-modal__header button')[0].click()

    if browser.find_by_id('post-title-0'):
        browser.find_by_id('post-title-0').fill(title)
    else:
        browser.find_by_css('.editor-post-title').first.type(title)

    browser.find_by_css('.editor-post-publish-button__button')[0].click()
    functional.eventually(browser.find_by_css, ['.editor-post-publish-button'])
    browser.find_by_css('.editor-post-publish-button')[0].click()


def _delete_post(browser, title):
    """Delete a blog post in WordPress site."""
    post = _get_post(browser, title)
    if not post:
        raise Exception('Post not found')

    delete_element = post.find_by_css('.submitdelete')[0]
    browser.visit(delete_element['href'])


def _get_post(browser, title):
    """Return whether a blog post with a given title is available."""
    _visit_site(browser)
    functional.visit(browser, '/wordpress/wp-admin/edit.php')
    xpath = '//tr[contains(@class, "type-post")][.//a[contains(@class, ' \
        f'"row-title") and contains(text(), "{title}")]]'
    post = browser.find_by_xpath(xpath)
    return post[0] if post else None


def _enable_public_mode(browser, should_enable):
    """Enable/disable the public mode."""
    checkbox = browser.find_by_id('id_is_public')
    if should_enable:
        checkbox.check()
    else:
        checkbox.uncheck()

    functional.submit(browser, form_class='form-configuration')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manifest for Shadowsocks Client.
"""

backup = {
    'secrets': {
        'files': [
            '/var/lib/private/shadowsocks-libev/freedombox/freedombox.json'
        ]
    },
    'services': ['shadowsocks-libev-local@freedombox']
}



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Views for the calibre module."""

from django.contrib import messages
from django.contrib.messages.views import SuccessMessageMixin
from django.http import Http404
from django.shortcuts import redirect
from django.template.response import TemplateResponse
from django.urls import reverse_lazy
from django.utils.translation import gettext as _
from django.views.generic.edit import FormView

from plinth import app as app_module
from plinth import views

from . import forms, privileged


class CalibreAppView(views.AppView):
    """Serve configuration form."""

    app_id = 'calibre'
    template_name = 'calibre.html'

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['libraries'] = privileged.list_libraries()
        return context


class CreateLibraryView(SuccessMessageMixin, FormView):
    """View to create an empty library."""

    form_class = forms.CreateLibraryForm
    prefix = 'calibre'
    template_name = 'form.html'
    success_url = reverse_lazy('calibre:index')
    success_message = _('Library created.')

    def form_valid(self, form):
        """Create the library on valid form submission."""
        try:
            privileged.create_library(form.cleaned_data['name'])
        except Exception as error:
            self.success_message = ''
            messages.error(
                self.request, "{0} {1}".format(
                    _('An error occurred while creating the library.'),
                    str(error)))

        return super().form_valid(form)


def delete_library(request, name):
    """View to delete a library."""
    if name not in privileged.list_libraries():
        raise Http404

    if request.method == 'POST':
        try:
            privileged.delete_library(name)
            messages.success(request, _('{name} deleted.').format(name=name))
        except Exception as error:
            messages.error(
                request,
                _('Could not delete {name}: {error}').format(
                    name=name, error=error))
        return redirect(reverse_lazy('calibre:index'))

    return TemplateResponse(request, 'calibre-delete-library.html', {
        'title': app_module.App.get('calibre').info.name,
        'name': name
    })



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2018-01-29 10:21
from __future__ import unicode_literals

import django.db.models.deletion
from django.conf import settings
from django.contrib.auth.models import User
from django.db import migrations, models

from plinth.models import UserProfile


def insert_users(apps, schema_editor):
    """For each user, create their empty profiles."""
    for user in User.objects.all():
        UserProfile(user=user).save()


def truncate_user_profile(apps, schema_editor):
    """Delete all user profiles."""
    UserProfile.objects.all().delete()


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('plinth', '0003_merge_firstboot_completed_fields'),
    ]

    operations = [
        migrations.CreateModel(
            name='UserProfile',
            fields=[
                ('id',
                 models.AutoField(auto_created=True, primary_key=True,
                                  serialize=False, verbose_name='ID')),
                ('language',
                 models.CharField(default=None, max_length=32, null=True)),
                ('user',
                 models.OneToOneField(
                     on_delete=django.db.models.deletion.CASCADE,
                     to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.RunPython(code=insert_users,
                             reverse_code=truncate_user_profile),
    ]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to configure OpenVPN server."""

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import cfg, frontpage, menu
from plinth.daemon import Daemon
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import Firewall
from plinth.modules.users.components import UsersAndGroups
from plinth.package import Packages
from plinth.utils import format_lazy

from . import manifest, privileged

_description = [
    format_lazy(
        _('Virtual Private Network (VPN) is a technique for securely '
          'connecting two devices in order to access resources of a '
          'private network.  While you are away from home, you can connect '
          'to your {box_name} in order to join your home network and '
          'access private/internal services provided by {box_name}. '
          'You can also access the rest of the Internet via {box_name} '
          'for added security and anonymity.'), box_name=_(cfg.box_name))
]

SERVER_CONFIGURATION_FILE = '/etc/openvpn/server/freedombox.conf'


class OpenVPNApp(app_module.App):
    """FreedomBox app for OpenVPN."""

    app_id = 'openvpn'

    _version = 5

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        self.groups = {'vpn': _('Connect to VPN services')}

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               name=_('OpenVPN'), icon_filename='openvpn',
                               short_description=_('Virtual Private Network'),
                               description=_description, manual_page='OpenVPN',
                               clients=manifest.clients)
        self.add(info)

        menu_item = menu.Menu('menu-openvpn', info.name,
                              info.short_description, info.icon_filename,
                              'openvpn:index', parent_url_name='apps')
        self.add(menu_item)

        download_profile = \
            format_lazy(_('<a class="btn btn-primary btn-sm" href="{link}">'
                          'Download Profile</a>'),
                        link=reverse_lazy('openvpn:profile'))
        shortcut = frontpage.Shortcut(
            'shortcut-openvpn', info.name,
            short_description=info.short_description, icon=info.icon_filename,
            description=info.description + [download_profile],
            manual_page=info.manual_page,
            configure_url=reverse_lazy('openvpn:index'), login_required=True,
            allowed_groups=['vpn'])
        self.add(shortcut)

        packages = Packages('packages-openvpn', ['openvpn', 'easy-rsa'])
        self.add(packages)

        firewall = Firewall('firewall-openvpn', info.name, ports=['openvpn'],
                            is_external=True)
        self.add(firewall)

        daemon = Daemon('daemon-openvpn', 'openvpn-server@freedombox',
                        listen_ports=[(1194, 'udp4'), (1194, 'udp6')])
        self.add(daemon)

        users_and_groups = UsersAndGroups('users-and-groups-openvpn',
                                          groups=self.groups)
        self.add(users_and_groups)

        backup_restore = BackupRestore('backup-restore-openvpn',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup()
        if not old_version:
            self.enable()

    def uninstall(self):
        """De-configure and uninstall the app."""
        super().uninstall()
        privileged.uninstall()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the Shadowsocks Clients.
"""

from django.urls import re_path

from .views import ShadowsocksAppView

urlpatterns = [
    re_path(r'^apps/shadowsocks/$', ShadowsocksAppView.as_view(),
            name='index'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.utils.translation import gettext_lazy as _

from plinth.clients import store_url

clients = [{
    'name':
        _('Mumble'),
    'platforms': [{
        'type': 'download',
        'os': 'gnu-linux',
        'url': 'https://wiki.mumble.info/wiki/Main_Page'
    }, {
        'type': 'download',
        'os': 'macos',
        'url': 'https://wiki.mumble.info/wiki/Main_Page'
    }, {
        'type': 'download',
        'os': 'windows',
        'url': 'https://wiki.mumble.info/wiki/Main_Page'
    }, {
        'type': 'package',
        'format': 'deb',
        'name': 'mumble'
    }, {
        'type': 'store',
        'os': 'ios',
        'store_name': 'app-store',
        'url': 'https://apps.apple.com/us/app/mumble/id443472808'
    }]
}, {
    'name':
        _('Mumblefly'),
    'platforms': [{
        'type': 'store',
        'os': 'ios',
        'store_name': 'app-store',
        'url': 'https://apps.apple.com/dk/app/mumblefy/id858752232'
    }]
}, {
    'name':
        _('Mumla'),
    'platforms': [{
        'type': 'store',
        'os': 'android',
        'store_name': 'f-droid',
        'url': store_url('f-droid', 'se.lublin.mumla')
    }, {
        'type': 'store',
        'os': 'android',
        'store_name': 'google-play',
        'url': store_url('google-play', 'se.lublin.mumla')
    }]
}]

backup = {
    'config': {
        'files': ['/etc/mumble-server.ini']
    },
    'data': {
        'directories': ['/var/lib/mumble-server']
    },
    'services': ['mumble-server']
}



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Module to handle glib main loop and provide asynchronous utilities.
"""

import logging
import random
import threading

from plinth import dbus, network
from plinth.utils import import_from_gi

from . import cfg

glib = import_from_gi('GLib', '2.0')

_thread = None
_main_loop = None

logger = logging.getLogger(__name__)


def run():
    """Run a glib main loop forever in a thread."""
    global _thread
    _thread = threading.Thread(target=_run)
    _thread.start()


def stop():
    """Exit glib main loop and end the thread."""
    if _main_loop:
        logger.info('Exiting glib main loop')
        _main_loop.quit()


def _run():
    """Connect to D-Bus and run main loop."""
    logger.info('Started new thread for glib main loop.')

    # Initialize all modules that use glib main loop
    dbus.init()
    network.init()

    global _main_loop
    _main_loop = glib.MainLoop()
    _main_loop.run()
    _main_loop = None

    logger.info('Glib main loop thread exited.')


def schedule(interval, method, data=None, in_thread=True, repeat=True,
             add_jitter=True):
    """Schedule a recurring call to a method with fixed interval."""

    def _runner():
        """Run the target method and log and exceptions."""
        try:
            return method(data)
        except Exception as exception:  # pylint: disable=broad-except
            logger.exception('Exception in running scheduled method - %s',
                             exception)

    def _run_bare_or_thread(_user_data):
        """Run the target method in thread or directly (if async)."""
        if not in_thread:
            _runner()
            return repeat

        thread = threading.Thread(target=_runner)
        thread.start()
        return repeat

    # When running in development mode, reduce the interval for tasks so that
    # they are triggered quickly and frequently to facilitate debugging.
    if cfg.develop and interval > 180:
        interval = 180

    if add_jitter:
        # Add or subtract 5% random jitter to given interval to avoid many
        # tasks running at exactly the same time (and competing for DB, disk,
        # network, etc.).
        interval *= 0.95 + (random.random() * 0.1)

    glib.timeout_add(int(interval * 1000), _run_bare_or_thread, None)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configuration helper for FreedomBox firewall interface."""

import subprocess
from typing import TypeAlias

import augeas

from plinth import action_utils
from plinth.actions import privileged

FirewallConfig: TypeAlias = dict[str, str | list[str]]


def _flush_iptables_rules():
    """Flush firewalld iptables rules before restarting it.

    This is workaround for
    https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=914694

    This workaround can be removed if the bug is fixed or if firewalld starts
    defaulting to nftables again.

    The bug leads to firewalld failing to flush rules when there are custom
    chains in the rules. This only happens on firewalld iptables backend when
    nftables is running with iptables compatibility.

    Flushing the tables before a restart will make the restart succeed and
    after the restart nftables backend is used avoiding the problem.

    """
    rule_template = '*{table}\n-F\n-X\n-Z\nCOMMIT\n'
    iptables_rules = ''
    ip6tables_rules = ''
    for table in ['security', 'raw', 'mangle', 'nat', 'filter']:
        iptables_rules += rule_template.format(table=table)
        ip6tables_rules += rule_template.format(table=table)

    subprocess.run(['iptables-restore'], input=iptables_rules.encode(),
                   check=True)
    subprocess.run(['ip6tables-restore'], input=iptables_rules.encode(),
                   check=True)


def set_firewall_backend(backend):
    """Set FirewallBackend attribute to the specified string."""
    conf_file = '/etc/firewalld/firewalld.conf'
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)

    # lens for shell-script config file
    aug.set('/augeas/load/Shellvars/lens', 'Shellvars.lns')
    aug.set('/augeas/load/Shellvars/incl[last() + 1]', conf_file)
    aug.load()

    old_backend = aug.get('/files/{}/FirewallBackend'.format(conf_file))
    aug.set('/files/{}/FirewallBackend'.format(conf_file),
            '{}'.format(backend))
    aug.save()

    if old_backend == 'iptables':
        _flush_iptables_rules()

    if backend != old_backend:
        action_utils.service_restart('firewalld')


def _setup_local_service_protection():
    """Create the basic set of direct rules for protecting local services.

    Local service protection means that only administrators and Apache web
    server should be able to access certain services and not other users who
    have logged into the system. This is needed because some of the services
    are protected with authentication and authorization provided by Apache web
    server. If services are contacted directly then auth can be bypassed by all
    local users.

    Firewalld does not have a mechanism to do this directly but it allows
    inserting 'direct' rules into firewall. nftables is our default backend by
    'direct' rules always invoke 'ip(6)tables' commands. Luckily, ip(6)tables
    are compatibility wrappers provided by nftables. Hence we must use iptables
    syntax even though we deal with nftables.

    In nftables, there is no direct way to write the blocking rules. To deal
    with traffic for incoming services, we have to write the rules an 'input'
    chain. However, this chain does not have the information about the user who
    originated this traffic. Only the 'output' chain has this information. This
    may be fixed in the future. See:
    https://github.com/firewalld/firewalld/issues/725

    Our workaround for the situation is to mark the packets in the 'output'
    chain and then use that wmark in the 'input'. Since we have a fixed set of
    users want to allow, a single bit in the 32bit 'mark' property of the
    packet is sufficient.
    """

    def _run_firewall_cmd(args):
        subprocess.run(args, stdout=subprocess.DEVNULL,
                       stderr=subprocess.DEVNULL, check=True)

    def _add_rule(permanent, *rule):
        try:
            _run_firewall_cmd(['firewall-cmd'] + permanent +
                              ['--direct', '--query-passthrough'] + list(rule))
        except subprocess.CalledProcessError:
            _run_firewall_cmd(['firewall-cmd'] + permanent +
                              ['--direct', '--add-passthrough'] + list(rule))

    for permanent in [[], ['--permanent']]:
        for ip_type in ['ipv4', 'ipv6']:
            for owner_type in ['--uid-owner', '--gid-owner']:
                for user_group in ['root', 'www-data']:
                    _add_rule(permanent, ip_type, '-A', 'OUTPUT', '-m',
                              'owner', owner_type, user_group, '-j', 'MARK',
                              '--or-mark', '0x800000')

    for permanent in [[], ['--permanent']]:
        for ip_type in ['ipv4', 'ipv6']:
            _add_rule(permanent, ip_type, '-A', 'INPUT', '-m', 'conntrack',
                      '--ctstate', 'ESTABLISHED,RELATED', '-j', 'ACCEPT')
            _add_rule(permanent, ip_type, '-A', 'INPUT', '-m', 'mark',
                      '--mark', '0x800000/0x800000', '-j', 'ACCEPT')


@privileged
def setup():
    """Perform basic firewalld setup."""
    action_utils.service_enable('firewalld')
    subprocess.run(['firewall-cmd', '--set-default-zone=external'],
                   stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL,
                   check=True)
    set_firewall_backend('nftables')

    _setup_local_service_protection()


@privileged
def get_config() -> FirewallConfig:
    """Return firewalld configuration for diagnostics."""
    config: FirewallConfig = {}

    # Get the default zone.
    output = subprocess.check_output(['firewall-cmd', '--get-default-zone'])
    config['default_zone'] = output.decode().strip()

    # Load Augeas lens.
    conf_file = '/etc/firewalld/firewalld.conf'
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.transform('Shellvars', conf_file)
    aug.set('/augeas/context', '/files' + conf_file)
    aug.load()

    # Get the firewall backend.
    config['backend'] = aug.get('FirewallBackend')

    # Get the list of direct passthroughs.
    output = subprocess.check_output(
        ['firewall-cmd', '--direct', '--get-all-passthroughs'])
    config['passthroughs'] = output.decode().strip().split('\n')

    return config



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure Syncthing."""

import grp
import os
import pwd
import shutil
import subprocess
import time

import augeas

from plinth import action_utils
from plinth.actions import privileged

DATA_DIR = '/var/lib/syncthing'
CONF_FILE = DATA_DIR + '/.config/syncthing/config.xml'


def augeas_load():
    """Initialize Augeas."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.add_transform('Xml.lns', CONF_FILE)
    aug.load()
    return aug


@privileged
def setup():
    """Perform post-install actions for Syncthing."""
    # Create syncthing group if needed.
    try:
        grp.getgrnam('syncthing')
    except KeyError:
        subprocess.run(['addgroup', '--system', 'syncthing'], check=True)

    # Create syncthing user if needed.
    try:
        pwd.getpwnam('syncthing')
    except KeyError:
        subprocess.run([
            'adduser', '--system', '--ingroup', 'syncthing', '--home',
            DATA_DIR, '--gecos', 'Syncthing file synchronization server',
            'syncthing'
        ], check=True)

    if not os.path.exists(DATA_DIR):
        os.makedirs(DATA_DIR, mode=0o750)
        shutil.chown(DATA_DIR, user='syncthing', group='syncthing')


@privileged
def setup_config():
    """Make configuration changes."""
    # wait until the configuration file is created by the syncthing daemon
    timeout = 300
    while timeout > 0:
        if os.path.exists(CONF_FILE):
            break
        timeout = timeout - 1
        time.sleep(1)

    aug = augeas_load()

    # disable authentication missing notification as FreedomBox itself
    # provides authentication
    auth_conf = ('/configuration/options/unackedNotificationID'
                 '[#text="authenticationUserAndPassword"]')
    conf_changed = bool(aug.remove('/files' + CONF_FILE + auth_conf))

    # disable usage reporting notification by declining reporting
    # if the user has not made a choice yet
    usage_conf = '/configuration/options/urAccepted/#text'
    if aug.get('/files' + CONF_FILE + usage_conf) == '0':
        aug.set('/files' + CONF_FILE + usage_conf, '-1')
        conf_changed = True

    aug.save()

    if conf_changed:
        action_utils.service_try_restart('syncthing@syncthing')


@privileged
def uninstall():
    """Remove configuration file when app is uninstalled."""
    shutil.rmtree(DATA_DIR + '/.config', ignore_errors=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Views for mumble app.
"""

from django.contrib import messages
from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth.modules import mumble
from plinth.modules.mumble.forms import MumbleForm
from plinth.views import AppView

from . import privileged


class MumbleAppView(AppView):
    app_id = 'mumble'
    form_class = MumbleForm

    def get_initial(self):
        """Return the values to fill in the form."""
        initial = super().get_initial()
        initial['domain'] = mumble.get_domain()
        initial['root_channel_name'] = privileged.get_root_channel_name()
        return initial

    def form_valid(self, form):
        """Apply form changes."""
        old_config = self.get_initial()
        new_config = form.cleaned_data

        if mumble.get_domain() != new_config['domain']:
            privileged.set_domain(new_config['domain'])
            app = app_module.App.get('mumble')
            app.get_component('letsencrypt-mumble').setup_certificates()
            messages.success(self.request, _('Configuration updated'))

        password = new_config.get('super_user_password')
        if password:
            privileged.set_super_user_password(password)
            messages.success(self.request,
                             _('SuperUser password successfully updated.'))

        join_password = new_config.get('join_password')
        if join_password:
            privileged.change_join_password(join_password)
            messages.success(self.request, _('Join password changed'))

        name = new_config.get('root_channel_name')
        if old_config['root_channel_name'] != new_config['root_channel_name']:
            privileged.change_root_channel_name(name)
            messages.success(self.request, _('Root channel name changed.'))

        return super().form_valid(form)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Views for the Matrix Synapse module."""

import datetime

from django.contrib import messages
from django.shortcuts import redirect
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.views.generic import FormView

from plinth import app as app_module
from plinth.forms import DomainSelectionForm
from plinth.modules import matrixsynapse, names
from plinth.modules.coturn.components import TurnConfiguration
from plinth.views import AppView

from . import get_turn_configuration, privileged
from .forms import MatrixSynapseForm


class SetupView(FormView):
    """Show matrix-synapse setup page."""

    template_name = 'matrix-synapse-pre-setup.html'
    form_class = DomainSelectionForm
    success_url = reverse_lazy('matrixsynapse:index')

    def form_valid(self, form):
        """Handle valid form submission."""
        matrixsynapse.setup_domain(form.cleaned_data['domain_name'])
        return super().form_valid(form)

    def get_context_data(self, *args, **kwargs):
        """Provide context data to the template."""
        context = super().get_context_data(**kwargs)
        app = app_module.App.get('matrixsynapse')

        context['title'] = app.info.name
        context['app_info'] = app.info
        context['domain_names'] = names.components.DomainName.list_names(
            'matrix-synapse-plinth')

        return context


class MatrixSynapseAppView(AppView):
    """Show matrix-synapse service page."""

    app_id = 'matrixsynapse'
    template_name = 'matrix-synapse.html'
    form_class = MatrixSynapseForm

    def dispatch(self, request, *args, **kwargs):
        """Redirect to setup page if setup is not done yet."""
        status = self.get_common_status()
        if status['is_enabled']:
            # App is disabled when uninstalling
            if not matrixsynapse.is_setup():
                return redirect('matrixsynapse:setup')

        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, *args, **kwargs):
        """Add additional context data for template."""
        context = super().get_context_data(*args, **kwargs)
        context['domain_name'] = matrixsynapse.get_configured_domain_name()
        context['certificate_status'] = matrixsynapse.get_certificate_status()
        context['config'] = privileged.get_config()
        tokens = privileged.list_registration_tokens()
        for token in tokens:
            if token['expiry_time']:
                date = datetime.datetime.utcfromtimestamp(
                    token['expiry_time'] / 1000)
                token['expiry_time'] = date

        context['registration_tokens'] = tokens
        return context

    def get_initial(self):
        """Return the values to fill in the form."""
        initial = super().get_initial()
        turn_config, managed = get_turn_configuration()
        config = privileged.get_config()
        initial.update({
            'enable_public_registration':
                config['public_registration'],
            'registration_verification':
                config['registration_verification'] or 'disabled',
            'enable_managed_turn':
                managed,
            'turn_uris':
                '\n'.join(turn_config.uris),
            'shared_secret':
                turn_config.shared_secret
        })
        return initial

    @staticmethod
    def _handle_turn_configuration(old_config, new_config):
        if not new_config['enable_managed_turn']:
            new_turn_uris = new_config['turn_uris'].splitlines()
            new_shared_secret = new_config['shared_secret']

            turn_config_changed = \
                old_config['turn_uris'] != new_turn_uris or \
                old_config['shared_secret'] != new_shared_secret

            if turn_config_changed:
                matrixsynapse.update_turn_configuration(
                    TurnConfiguration(None, new_turn_uris, new_shared_secret),
                    managed=False)
        else:
            # Remove overridden turn configuration
            matrixsynapse.update_turn_configuration(TurnConfiguration(),
                                                    managed=False)

    def form_valid(self, form):
        """Handle valid form submission."""
        old_config = self.get_initial()
        new_config = form.cleaned_data

        def changed(prop):
            return old_config[prop] != new_config[prop]

        is_changed = False
        if (changed('enable_public_registration')
                or changed('registration_verification')):
            try:
                privileged.set_config(
                    public_registration=new_config[
                        'enable_public_registration'],
                    registration_verification=new_config[
                        'registration_verification'])
                is_changed = True
            except ProcessLookupError:
                # Matrix Synapse server is not running
                messages.error(
                    self.request,
                    _('Registration configuration cannot be updated when app '
                      'is disabled.'))

        if changed('enable_managed_turn') or changed('turn_uris') or \
           changed('shared_secret'):
            self._handle_turn_configuration(old_config, new_config)
            is_changed = True

        if is_changed:
            messages.success(self.request, _('Configuration updated'))

        return super().form_valid(form)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the power module.
"""

from django.urls import re_path

from . import views

urlpatterns = [
    re_path(r'^sys/power/$', views.PowerAppView.as_view(), name='index'),
    re_path(r'^sys/power/restart$', views.restart, name='restart'),
    re_path(r'^sys/power/shutdown$', views.shutdown, name='shutdown'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Diagnostic check data type."""

import dataclasses
import json
from dataclasses import dataclass, field
from enum import StrEnum
from typing import TypeAlias

from django.utils.translation import gettext

from plinth.utils import SafeFormatter

DiagnosticCheckParameters: TypeAlias = dict[str, str | int | bool | None]


class Result(StrEnum):
    """The result of a diagnostic check."""
    NOT_DONE = 'not_done'
    PASSED = 'passed'
    WARNING = 'warning'
    FAILED = 'failed'
    ERROR = 'error'


@dataclass
class DiagnosticCheck:
    """A diagnostic check and optional result and parameters."""
    check_id: str
    description: str
    result: Result = Result.NOT_DONE
    parameters: DiagnosticCheckParameters = field(default_factory=dict)
    component_id: str | None = None

    @property
    def translated_description(self):
        """Return translated string for description."""
        description = gettext(self.description)
        if self.parameters:
            return SafeFormatter().vformat(description, [], self.parameters)

        return description


class CheckJSONEncoder(json.JSONEncoder):
    """Encode objects that include DiagnosticChecks."""

    def default(self, o):
        """Add class tag to DiagnosticChecks."""
        if isinstance(o, DiagnosticCheck):
            o = dataclasses.asdict(o)
            o.update({'__class__': 'DiagnosticCheck'})
            return o

        return super().default(o)


class CheckJSONDecoder(json.JSONDecoder):
    """Decode objects that include DiagnosticChecks."""

    def __init__(self):
        json.JSONDecoder.__init__(self, object_hook=CheckJSONDecoder.from_dict)

    @staticmethod
    def from_dict(data):
        """Convert tagged data to DiagnosticCheck."""
        if data.get('__class__') == 'DiagnosticCheck':
            return DiagnosticCheck(data['check_id'], data['description'],
                                   data['result'], data['parameters'],
                                   data.get('component_id'))

        return data



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for tor app.
"""

import pytest

from plinth.tests import functional

_TOR_FEATURE_TO_ELEMENT = {
    'relay': 'tor-relay_enabled',
    'bridge-relay': 'tor-bridge_relay_enabled',
    'hidden-services': 'tor-hs_enabled',
}

pytestmark = [pytest.mark.apps, pytest.mark.domain, pytest.mark.tor]


class TestTorApp(functional.BaseAppTests):
    app_name = 'tor'
    has_service = True
    has_web = False

    def test_set_tor_relay_configuration(self, session_browser):
        """Test setting Tor relay configuration."""
        functional.app_enable(session_browser, 'tor')
        _feature_enable(session_browser, 'relay', should_enable=False)
        _feature_enable(session_browser, 'relay', should_enable=True)
        _assert_feature_enabled(session_browser, 'relay', enabled=True)
        _assert_relay_port(session_browser, 'orport')

    def test_set_tor_bridge_relay_configuration(self, session_browser):
        """Test setting Tor bridge relay configuration."""
        functional.app_enable(session_browser, 'tor')
        _feature_enable(session_browser, 'bridge-relay', should_enable=False)
        _feature_enable(session_browser, 'bridge-relay', should_enable=True)
        _assert_feature_enabled(session_browser, 'bridge-relay', enabled=True)
        _assert_relay_port(session_browser, 'obfs3')
        _assert_relay_port(session_browser, 'obfs4')

    def test_set_tor_hidden_services_configuration(self, session_browser):
        """Test setting Tor hidden services configuration."""
        functional.app_enable(session_browser, 'tor')
        _feature_enable(session_browser, 'hidden-services',
                        should_enable=False)
        _feature_enable(session_browser, 'hidden-services', should_enable=True)
        _assert_feature_enabled(session_browser, 'hidden-services',
                                enabled=True)
        _assert_hidden_services(session_browser)

    # TODO: Test more thoroughly by checking same hidden service is restored
    # and by actually connecting using Tor.
    @pytest.mark.backups
    def test_backup_restore(self, session_browser):
        """Test backup and restore of configuration."""
        functional.app_enable(session_browser, 'tor')
        _feature_enable(session_browser, 'relay', should_enable=True)
        _feature_enable(session_browser, 'bridge-relay', should_enable=True)
        _feature_enable(session_browser, 'hidden-services', should_enable=True)
        functional.backup_create(session_browser, 'tor', 'test_tor')

        _feature_enable(session_browser, 'relay', should_enable=False)
        _feature_enable(session_browser, 'hidden-services',
                        should_enable=False)
        functional.backup_restore(session_browser, 'tor', 'test_tor')

        assert functional.service_is_running(session_browser, 'tor')
        _assert_feature_enabled(session_browser, 'relay', enabled=True)
        _assert_feature_enabled(session_browser, 'bridge-relay', enabled=True)
        _assert_feature_enabled(session_browser, 'hidden-services',
                                enabled=True)


def _feature_enable(browser, feature, should_enable):
    """Enable/disable a Tor feature."""
    element_name = _TOR_FEATURE_TO_ELEMENT[feature]
    functional.nav_to_module(browser, 'tor')
    checkbox_element = browser.find_by_name(element_name).first
    if should_enable == checkbox_element.checked:
        return

    if should_enable:
        if feature == 'bridge-relay':
            browser.find_by_name('tor-relay_enabled').first.check()

        checkbox_element.check()
    else:
        checkbox_element.uncheck()

    functional.submit(browser, form_class='form-configuration')
    functional.wait_for_config_update(browser, 'tor')


def _assert_feature_enabled(browser, feature, enabled):
    """Assert whether Tor relay is enabled or disabled."""
    element_name = _TOR_FEATURE_TO_ELEMENT[feature]
    functional.nav_to_module(browser, 'tor')
    assert browser.find_by_name(element_name).first.checked == enabled


def _assert_relay_port(browser, port_name):
    """Assert that port is available in port forwarding info table."""
    functional.nav_to_module(browser, 'tor')
    assert f'tor-{port_name}' in [
        name.text for name in browser.find_by_css(
            '.table-port-forwarding-info td:first-child')
    ]


def _assert_hidden_services(browser):
    """Assert that hidden service information is shown."""
    functional.nav_to_module(browser, 'tor')
    assert browser.find_by_css('.tor-hs .tor-hs-hostname')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Tests for letsencrypt module.
"""

from unittest.mock import call, patch

import pytest

from plinth.modules.names.components import DomainType

from .. import on_domain_added, on_domain_removed


@pytest.fixture(name='domain_types')
def fixture_domain_types():
    """Create a domain types required for tests."""
    DomainType('domain-type-tor', 'Tor Onion Service', 'tor:index',
               can_have_certificate=False)
    DomainType('domain-type-test', 'Test Domain Type', 'test:index')


@pytest.mark.usefixtures('domain_types')
def test_add_onion_domain():
    """Test that .onion domains are ignored when added/removed."""
    assert not on_domain_added('test', 'domain-type-tor', 'ddddd.onion')
    assert not on_domain_removed('test', 'domain-type-tor', 'ddddd.onion')


@patch('plinth.modules.letsencrypt.get_status')
@patch('plinth.modules.letsencrypt.certificate_obtain')
@pytest.mark.usefixtures('load_cfg')
@pytest.mark.parametrize('domain,status_input,obtain,result', [
    ('domain1.tld', {
        'certificate_available': True,
        'validity': 'not-valid'
    }, True, True),
    ('domain2.tld', {
        'certificate_available': False,
        'validity': 'valid'
    }, True, True),
    ('domain3.tld', {
        'certificate_available': True,
        'validity': 'valid'
    }, False, False),
    ('', {
        'certificate_available': False,
        'validity': 'valid'
    }, False, True),
])
def test_add_valid_domain(certificate_obtain, get_status, domain, status_input,
                          obtain, result):
    """Test adding a domain that can have certificates."""
    get_status.return_value = {'domains': {domain: status_input}}
    assert result == on_domain_added('test', 'domain-type-test', domain)
    if obtain:
        certificate_obtain.assert_has_calls([call(domain)])
    else:
        certificate_obtain.assert_not_called()


@patch('plinth.modules.letsencrypt.certificate_revoke')
@pytest.mark.usefixtures('load_cfg')
@pytest.mark.parametrize('domain,revoke,result', [
    ('domain1.tld', True, True),
    ('', False, True),
])
def test_remove_domain(certificate_revoke, domain, revoke, result):
    """Test removing a domain that can certificates."""
    assert result == on_domain_removed('test', 'domain-type-test', domain)
    if revoke:
        certificate_revoke.assert_has_calls(
            [call(domain, really_revoke=False)])
    else:
        certificate_revoke.assert_not_called()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Forms for configuring Tor.
"""

import re

from django import forms
from django.core.exceptions import ValidationError
from django.core.validators import validate_ipv46_address
from django.forms import widgets
from django.utils.translation import gettext_lazy as _

from plinth import cfg
from plinth.utils import format_lazy


class TrimmedCharField(forms.CharField):
    """Trim the contents of a CharField"""

    def clean(self, value):
        """Clean and validate the field value"""
        if value:
            value = value.strip()
            value = value.replace('\r\n', '\n')

        return super().clean(value)


def bridges_validator(bridges):
    """Validate upstream bridges entries."""
    validation_error = ValidationError(
        _('Enter a valid bridge with this format: '
          '[transport] IP:ORPort [fingerprint]'), code='invalid')

    bridges = [bridge.strip() for bridge in bridges.split('\n')]
    bridges = [bridge for bridge in bridges if bridge]
    if not bridges:
        raise validation_error

    for bridge in bridges:
        parts = [part for part in bridge.split() if part]

        # IP:ORPort is required, transport and fingerprint are optional.
        # Transports may have additional options after the fingerprint.
        try:
            ip_port_part = parts[0]
            if re.match('[a-z_][a-z0-9_]*', parts[0]):
                ip_port_part = parts[1]
        except IndexError:
            raise validation_error

        match = re.match(r'\[([a-fA-F0-9:]+)\](?::([0-9]+))?', ip_port_part)
        if match:
            ip_address = match.group(1)
            port = match.group(2)
        else:
            ip_parts = ip_port_part.rsplit(':', maxsplit=1)
            ip_address = ip_parts[0]
            port = ip_parts[1] if len(ip_parts) > 1 else None

        try:
            validate_ipv46_address(ip_address)
        except ValidationError:
            raise validation_error

        if port:
            port = int(port)
            if port < 0 or port > 65535:
                raise validation_error


class TorCommonForm(forms.Form):
    """Tor common configuration form."""
    use_upstream_bridges = forms.BooleanField(
        label=_('Use upstream bridges to connect to Tor network'),
        required=False, help_text=_(
            'When enabled, the bridges configured below will be used to '
            'connect to the Tor network. Use this option if your Internet '
            'Service Provider (ISP) blocks or censors connections to the '
            'Tor Network. This will disable relay modes.'))
    upstream_bridges = TrimmedCharField(
        widget=widgets.Textarea, label=_('Upstream bridges'), required=False,
        help_text=_(
            'You can get some bridges from <a '
            'href="https://bridges.torproject.org/">'
            'https://bridges.torproject.org/</a> and copy/paste the bridge '
            'information here. Currently supported transports are none, '
            'obfs3, obfs4 and scamblesuit.'), validators=[bridges_validator])


class TorForm(TorCommonForm):
    """Tor configuration form."""
    relay_enabled = forms.BooleanField(
        label=_('Enable Tor relay'), required=False, help_text=format_lazy(
            _('When enabled, your {box_name} will run a Tor relay and donate '
              'bandwidth to the Tor network. Do this if you have more than '
              '2 megabits/s of upload and download bandwidth.'),
            box_name=_(cfg.box_name)))
    bridge_relay_enabled = forms.BooleanField(
        label=_('Enable Tor bridge relay'), required=False,
        help_text=format_lazy(
            _('When enabled, relay information is published in the Tor bridge '
              'database instead of public Tor relay database making it harder '
              'to censor this node. This helps others circumvent censorship.'),
            box_name=_(cfg.box_name)))
    hs_enabled = forms.BooleanField(
        label=_('Enable Tor Onion Service'), required=False,
        help_text=format_lazy(
            _('An onion service will allow {box_name} to provide selected '
              'services (such as wiki or chat) without revealing its '
              'location. Do not use this for strong anonymity yet.'),
            box_name=_(cfg.box_name)))

    def clean(self):
        """Validate the form for cross-field integrity."""
        cleaned_data = super().clean()
        use_upstream_bridges = cleaned_data.get('use_upstream_bridges')
        upstream_bridges = cleaned_data.get('upstream_bridges')

        if use_upstream_bridges and not upstream_bridges:
            self.add_error(
                'upstream_bridges',
                ValidationError(
                    _('Specify at least one upstream bridge to use upstream '
                      'bridges.'), code='invalid'))

        return cleaned_data



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for matrixsynapse app.
"""

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.matrixsynapse]


class TestMatrixSynapseApp(functional.BaseAppTests):
    app_name = 'matrixsynapse'
    has_service = True
    has_web = False
    diagnostics_delay = 1

    @pytest.fixture(scope='class', autouse=True)
    def fixture_setup(self, session_browser):
        """Setup the app."""
        functional.login(session_browser)
        functional.set_domain_name(session_browser, 'mydomain.example')

    def install_and_setup(self, session_browser):
        """Install the app and run setup."""
        super().install_and_setup(session_browser)
        functional.app_select_domain_name(session_browser, self.app_name,
                                          'mydomain.example')

    def test_uninstall(self, session_browser):
        """After uninstall test, after installing select the domain again."""
        super().test_uninstall(session_browser)
        functional.app_select_domain_name(session_browser, self.app_name,
                                          'mydomain.example')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Views for the sharing app."""

from django.contrib import messages
from django.contrib.messages.views import SuccessMessageMixin
from django.http import Http404
from django.shortcuts import redirect
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.views.decorators.http import require_POST
from django.views.generic import FormView

from plinth.views import AppView

from . import privileged
from .forms import AddShareForm


class SharingAppView(AppView):
    """Sharing configuration page."""

    app_id = 'sharing'
    template_name = 'sharing.html'

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['shares'] = privileged.list_shares()
        return context


class AddShareView(SuccessMessageMixin, FormView):
    """View to add a new share."""

    form_class = AddShareForm
    prefix = 'sharing'
    template_name = 'sharing_add_edit.html'
    success_url = reverse_lazy('sharing:index')
    success_message = _('Share added.')

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Add Share')
        return context

    def form_valid(self, form):
        """Add the share on valid form submission."""
        _add_share(form.cleaned_data)
        return super().form_valid(form)


class EditShareView(SuccessMessageMixin, FormView):
    """View to edit an existing share."""

    form_class = AddShareForm
    prefix = 'sharing'
    template_name = 'sharing_add_edit.html'
    success_url = reverse_lazy('sharing:index')
    success_message = _('Share edited.')

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Edit Share')
        return context

    def get_initial(self):
        """Load information about share being edited."""
        try:
            return [
                share for share in privileged.list_shares()
                if share['name'] == self.kwargs['name']
            ][0]
        except IndexError:
            raise Http404

    def form_valid(self, form):
        """Add the share on valid form submission."""
        if form.initial != form.cleaned_data:
            privileged.remove(form.initial['name'])
            _add_share(form.cleaned_data)

        return super().form_valid(form)


def _add_share(form_data):
    privileged.add(form_data['name'], form_data['path'], form_data['groups'],
                   form_data['is_public'])


@require_POST
def remove(request, name):
    """View to remove a share."""
    privileged.remove(name)
    messages.success(request, _('Share deleted.'))
    return redirect(reverse_lazy('sharing:index'))



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Django views for Tiny Tiny RSS app."""

from django.contrib import messages
from django.utils.translation import gettext_lazy as _

from plinth.forms import TLSDomainForm
from plinth.views import AppView

from . import privileged


class TTRSSAppView(AppView):
    """Show TTRSS app main view."""

    app_id = 'ttrss'
    form_class = TLSDomainForm

    def get_initial(self):
        """Return the values to fill in the form."""
        initial = super().get_initial()
        initial['domain'] = privileged.get_domain()
        return initial

    def form_valid(self, form):
        """Change the domain of TT-RSS app."""
        data = form.cleaned_data
        old_data = form.initial
        if old_data['domain'] != data['domain']:
            privileged.set_domain(data['domain'])
            messages.success(self.request, _('Configuration updated'))

        return super().form_valid(form)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test module for ejabberd STUN/TURN configuration.
"""

import pathlib
import shutil
from unittest.mock import Mock, patch

import pytest

from plinth.modules import ejabberd
from plinth.modules.coturn.components import TurnConfiguration
from plinth.modules.ejabberd import privileged

managed_configuration = TurnConfiguration(
    'freedombox.local', [],
    'aiP02OAGyOlj6WGuCyqj7iaOsbkC7BUeKvKzhAsTZ8MEwMd3yTwpr2uvbOxgWe51')

overridden_configuration = TurnConfiguration(
    'public.coturn.site', [],
    'BUeKvKzhAsTZ8MEwMd3yTwpr2uvbOxgWe51aiP02OAGyOlj6WGuCyqj7iaOsbkC7')

pytestmark = pytest.mark.usefixtures('mock_privileged')
current_directory = pathlib.Path(__file__).parent
privileged_modules_to_mock = ['plinth.modules.ejabberd.privileged']


@pytest.fixture(name='conf_file')
def fixture_conf_file(tmp_path):
    """Uses a dummy configuration file."""
    settings_file_name = 'ejabberd.yml'
    conf_file = tmp_path / settings_file_name
    conf_file.touch()
    shutil.copyfile(str(current_directory / 'data' / 'ejabberd.yml.example'),
                    str(conf_file))
    return str(conf_file)


@pytest.fixture(name='managed_file')
def fixture_managed_file(tmp_path):
    """Uses a dummy managed file."""
    file_name = 'freedombox_managed_coturn'
    fil = tmp_path / file_name
    return str(fil)


@pytest.fixture(autouse=True)
def fixture_set_configuration_paths(conf_file, managed_file):
    """Setup configuration file paths in action module."""
    privileged.EJABBERD_CONFIG = conf_file
    privileged.EJABBERD_MANAGED_COTURN = managed_file


@pytest.fixture(name='test_configuration', autouse=True)
def fixture_test_configuration(conf_file):
    """Use a separate ejabberd configuration for tests.

    The module state is patched to be 'up-to-date'.
    """
    with patch('plinth.app.App.get') as app_get:
        app = Mock()
        app_get.return_value = app
        app.needs_setup.return_value = False
        yield


def _set_turn_configuration(config=managed_configuration, managed=True):
    with patch('plinth.action_utils.service_is_running', return_value=False):
        ejabberd.update_turn_configuration(config, managed=managed)


def _assert_conf(expected_configuration, expected_managed):
    """Assert that ejabberd TURN configuration is as expected."""
    config, managed = ejabberd.get_turn_configuration()
    assert config.uris == expected_configuration.uris
    assert config.shared_secret == expected_configuration.shared_secret
    assert managed == expected_managed


def test_managed_turn_server_configuration():
    _set_turn_configuration()
    _assert_conf(managed_configuration, True)


def test_overridden_turn_server_configuration():
    _set_turn_configuration(overridden_configuration, False)
    _assert_conf(overridden_configuration, False)


def test_remove_turn_configuration():
    _set_turn_configuration()
    _set_turn_configuration(TurnConfiguration(), False)
    _assert_conf(TurnConfiguration(), False)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app for name services.
"""

from plinth.views import AppView

from . import components


class NamesAppView(AppView):
    """Show names app main page."""

    app_id = 'names'
    template_name = 'names.html'

    def get_context_data(self, *args, **kwargs):
        """Add additional context data for template."""
        context = super().get_context_data(*args, **kwargs)
        context['status'] = get_status()
        return context


def get_status():
    """Get configured services per name."""
    domains = components.DomainName.list()
    used_domain_types = {domain.domain_type for domain in domains}
    unused_domain_types = [
        domain_type for domain_type in components.DomainType.list().values()
        if domain_type not in used_domain_types
    ]

    return {'domains': domains, 'unused_domain_types': unused_domain_types}
